# Copyright (c) 2016 AT&T Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_config import fixture as config_fixture
from oslo_serialization import jsonutils
from webob import exc

from murano.api.v1 import environments
from murano.api.v1 import services
from murano.api.v1 import sessions
import murano.tests.unit.api.base as tb


class TestServicesApi(tb.ControllerTest, tb.MuranoApiTestCase):

    def setUp(self):
        super(TestServicesApi, self).setUp()
        self.environments_controller = environments.Controller()
        self.sessions_controller = sessions.Controller()
        self.services_controller = services.Controller()
        self.fixture = self.useFixture(config_fixture.Config())
        self.fixture.conf(args=[])

    def test_can_post(self):
        CREDENTIALS_1 = {'tenant': 'test_tenant_1', 'user': 'test_user_1'}
        self._set_policy_rules(
            {'create_environment': '@'}
        )
        self.expect_policy_check('create_environment')

        # Create environment
        request = self._post(
            '/environments',
            jsonutils.dump_as_bytes({'name': 'test_environment_1'}),
            **CREDENTIALS_1
        )
        response_body = jsonutils.loads(request.get_response(self.api).body)
        self.assertEqual(CREDENTIALS_1['tenant'],
                         response_body['tenant_id'])

        environment_id = response_body['id']

        # Create session
        request = self._post('/environments/{environment_id}/configure'
                             .format(environment_id=environment_id), b'',
                             **CREDENTIALS_1)
        response_body = jsonutils.loads(request.get_response(self.api).body)

        session_id = response_body['id']

        path = "/"

        request = self._post('/v1/environments/{0}/services'.
                             format(environment_id), b'',
                             **CREDENTIALS_1)
        request.headers['X-Configuration-Session'] = str(session_id)
        request.context.session = session_id

        self.assertRaises(exc.HTTPBadRequest, self.services_controller.post,
                          request, environment_id, path)

        response = self.services_controller.post(request, environment_id,
                                                 path, "test service")
        self.assertEqual("test service", response)

    def test_can_put(self):
        CREDENTIALS_1 = {'tenant': 'test_tenant_1', 'user': 'test_user_1'}
        self._set_policy_rules(
            {'create_environment': '@'}
        )
        self.expect_policy_check('create_environment')

        # Create environment
        request = self._post(
            '/environments',
            jsonutils.dump_as_bytes({'name': 'test_environment_1'}),
            **CREDENTIALS_1
        )
        response_body = jsonutils.loads(request.get_response(self.api).body)
        self.assertEqual(CREDENTIALS_1['tenant'],
                         response_body['tenant_id'])

        environment_id = response_body['id']

        # Create session
        request = self._post('/environments/{environment_id}/configure'
                             .format(environment_id=environment_id), b'',
                             **CREDENTIALS_1)
        response_body = jsonutils.loads(request.get_response(self.api).body)

        session_id = response_body['id']

        path = "/"

        request = self._put('/v1/environments/{0}/services'.
                            format(environment_id), b'',
                            **CREDENTIALS_1)
        request.headers['X-Configuration-Session'] = str(session_id)
        request.context.session = session_id

        # Check that empty body can be put
        response = self.services_controller.put(request, environment_id,
                                                path, [])
        self.assertEqual([], response)

        response = self.services_controller.put(request, environment_id,
                                                path, "test service")
        self.assertEqual("test service", response)

    def test_can_get(self):
        CREDENTIALS_1 = {'tenant': 'test_tenant_1', 'user': 'test_user_1'}
        self._set_policy_rules(
            {'create_environment': '@'}
        )
        self.expect_policy_check('create_environment')

        # Create environment
        request = self._post(
            '/environments',
            jsonutils.dump_as_bytes({'name': 'test_environment_1'}),
            **CREDENTIALS_1
        )
        response_body = jsonutils.loads(request.get_response(self.api).body)
        self.assertEqual(CREDENTIALS_1['tenant'],
                         response_body['tenant_id'])

        environment_id = response_body['id']

        # Create session
        request = self._post('/environments/{environment_id}/configure'
                             .format(environment_id=environment_id), b'',
                             **CREDENTIALS_1)
        response_body = jsonutils.loads(request.get_response(self.api).body)

        session_id = response_body['id']

        # Create service
        path = '/'
        request = self._post('/v1/environments/{0}/services'.
                             format(environment_id), b'',
                             **CREDENTIALS_1)
        request.headers['X-Configuration-Session'] = str(session_id)
        request.context.session = session_id

        response = self.services_controller.post(request, environment_id,
                                                 path, "test service")
        # Get service
        request = self._get('/v1/environments/{0}/services'.
                            format(environment_id), b'',
                            **CREDENTIALS_1)

        response = self.services_controller.get(request, environment_id, path)
        self.assertEqual([], response)

        request.headers['X-Configuration-Session'] = str(session_id)
        request.context.session = session_id

        response = self.services_controller.get(request, environment_id, path)
        self.assertNotEqual([], response)

    def test_can_delete(self):
        CREDENTIALS_1 = {'tenant': 'test_tenant_1', 'user': 'test_user_1'}
        self._set_policy_rules(
            {'create_environment': '@'}
        )
        self.expect_policy_check('create_environment')

        # Create environment
        request = self._post(
            '/environments',
            jsonutils.dump_as_bytes({'name': 'test_environment_1'}),
            **CREDENTIALS_1
        )
        response_body = jsonutils.loads(request.get_response(self.api).body)
        self.assertEqual(CREDENTIALS_1['tenant'],
                         response_body['tenant_id'])

        environment_id = response_body['id']

        # Create session
        request = self._post(
            '/environments/{environment_id}/configure'
            .format(environment_id=environment_id),
            b'', **CREDENTIALS_1
        )
        response_body = jsonutils.loads(request.get_response(self.api).body)

        session_id = response_body['id']

        # Create service
        path = '/'
        request = self._post('/v1/environments/{0}/services'.
                             format(environment_id), b'',
                             **CREDENTIALS_1)
        request.headers['X-Configuration-Session'] = str(session_id)
        request.context.session = session_id

        self.services_controller.post(request, environment_id,
                                      path, "test service")

        # Delete service
        request = self._delete('/v1/environments/{0}/services'.
                               format(environment_id), b'',
                               **CREDENTIALS_1)

        self.assertRaises(exc.HTTPBadRequest, self.services_controller.delete,
                          request, environment_id, path)

        request.headers['X-Configuration-Session'] = str(session_id)
        request.context.session = session_id

        self.assertRaises(exc.HTTPNotFound, self.services_controller.delete,
                          request, environment_id, path)



# Copyright 2016 AT&T Corp
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from murano.db import models
from murano.db.services import cf_connections
from murano.db.services import environments
from murano.db import session as db_session
from murano.tests.unit import base


class TestCFConnection(base.MuranoWithDBTestCase):

    def setUp(self):
        super(TestCFConnection, self).setUp()
        self.resources = []
        self.tenant_id = 'test_tenant_id'

        environment = models.Environment(
            name='test_environment',
            tenant_id='test_tenant_id',
            version=1
        )
        alt_environment = models.Environment(
            name='alt_test_environment',
            tenant_id='alt_test_tenant_id',
            version=1
        )

        unit = db_session.get_session()
        with unit.begin():
            unit.add_all([environment, alt_environment])

        self.environment_id = environment.id
        self.alt_environment_id = alt_environment.id

    def tearDown(self):
        super(TestCFConnection, self).tearDown()
        unit = db_session.get_session()

        if self.environment_id:
            environments.EnvironmentServices.remove(self.environment_id)
        if self.alt_environment_id:
            environments.EnvironmentServices.remove(self.alt_environment_id)

        for resource in self.resources:
            if resource:
                with unit.begin():
                    unit.delete(resource)

    def _create_resource(self, resource_name, **kwargs):
        unit = db_session.get_session()
        model = None
        new_resource_id = 1

        if resource_name == 'cf_org':
            model = models.CFOrganization
        elif resource_name == 'cf_space':
            model = models.CFSpace
        elif resource_name == 'cf_service':
            model = models.CFServiceInstance

        resource = unit.query(model).order_by(model.id.desc()).first()
        if resource:
            new_resource_id = int(resource.id) + 1
        new_resource = model(id=new_resource_id, **kwargs)

        with unit.begin():
            unit.add(new_resource)

        created_resource = unit.query(model)\
            .order_by(model.id.desc()).first()
        self.assertIsNotNone(created_resource)
        self.assertEqual(created_resource.id, str(new_resource.id))
        self.resources.append(created_resource)

        return created_resource

    def test_set_tenant_for_org(self):
        cf_org = self._create_resource('cf_org', tenant=self.tenant_id)
        cf_connections.set_tenant_for_org(cf_org.id, 'another_test_tenant_id')

        duplicate_cf_org = None
        unit = db_session.get_session()
        with unit.begin():
            duplicate_cf_org = unit.query(models.CFOrganization).get(cf_org.id)
        self.assertIsNotNone(duplicate_cf_org)
        self.assertEqual('another_test_tenant_id',
                         duplicate_cf_org.tenant)

    def test_set_environment_for_space(self):
        cf_space = self._create_resource('cf_space',
                                         environment_id=self.environment_id)
        cf_connections.set_environment_for_space(cf_space.id,
                                                 self.alt_environment_id)

        duplicate_cf_space = None
        unit = db_session.get_session()
        with unit.begin():
            duplicate_cf_space = unit.query(models.CFSpace).get(cf_space.id)
        self.assertIsNotNone(duplicate_cf_space)
        self.assertEqual(self.alt_environment_id,
                         duplicate_cf_space.environment_id)

    def test_set_instance_for_service(self):
        cf_service = self._create_resource('cf_service',
                                           service_id='123',
                                           environment_id=self.environment_id,
                                           tenant=self.tenant_id)
        cf_connections.set_instance_for_service(
            instance_id=cf_service.id, service_id='123',
            environment_id=self.alt_environment_id, tenant=self.tenant_id)

        duplicate_cf_service = None
        unit = db_session.get_session()
        with unit.begin():
            duplicate_cf_service = unit.query(models.CFServiceInstance).\
                get(cf_service.id)
        self.assertIsNotNone(duplicate_cf_service)
        self.assertEqual(self.alt_environment_id,
                         duplicate_cf_service.environment_id)

    def test_get_environment_for_space(self):
        cf_space = self._create_resource('cf_space',
                                         environment_id=self.environment_id)
        retrieved_env_id =\
            cf_connections.get_environment_for_space(cf_space.id)
        self.assertEqual(cf_space.environment_id,
                         retrieved_env_id)

    def test_get_tenant_for_org(self):
        cf_org = self._create_resource('cf_org', tenant=self.tenant_id)
        retrieved_tenant =\
            cf_connections.get_tenant_for_org(cf_org.id)
        self.assertEqual(self.tenant_id, retrieved_tenant)

    def test_get_service_for_instance(self):
        cf_service = self._create_resource('cf_service',
                                           service_id='123',
                                           environment_id=self.environment_id,
                                           tenant=self.tenant_id)
        retrieved_service =\
            cf_connections.get_service_for_instance(cf_service.id)
        for attr in ['id', 'environment_id', 'service_id', 'tenant']:
            self.assertEqual(getattr(cf_service, attr),
                             getattr(retrieved_service, attr))

    def test_delete_environment_from_space(self):
        environment = models.Environment(
            name='test_environment_1',
            tenant_id='test_tenant_id_1',
            version=1
        )
        unit = db_session.get_session()
        with unit.begin():
            unit.add(environment)

        cf_connections.delete_environment_from_space(environment.id)
        with unit.begin():
            retrieved_env = unit.query(models.Environment).get(environment.id)
            self.assertIsNotNone(retrieved_env)



#    Copyright (c) 2014 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


class InternalFlowException(Exception):
    pass


class ReturnException(InternalFlowException):
    def __init__(self, value):
        self._value = value

    @property
    def value(self):
        return self._value


class BreakException(InternalFlowException):
    pass


class ContinueException(InternalFlowException):
    pass


class DslInvalidOperationError(Exception):
    pass


class NoMethodFound(Exception):
    def __init__(self, name):
        super(NoMethodFound, self).__init__('Method "%s" is not found' % name)


class NoPropertyFound(Exception):
    def __init__(self, name):
        super(NoPropertyFound, self).__init__(
            'Property "%s" is not found' % name)


class NoClassFound(Exception):
    def __init__(self, name, packages=None):
        if packages is None:
            packages = []
        packages = ', '.join("{0}/{1}".format(p.name, p.version)
                             for p in packages)
        super(NoClassFound, self).__init__(
            'Class "{0}" is not found in {1}'.format(name, packages))


class NoPackageFound(Exception):
    def __init__(self, name):
        super(NoPackageFound, self).__init__(
            'Package "%s" is not found' % name)


class NoPackageForClassFound(Exception):
    def __init__(self, name):
        super(NoPackageForClassFound, self).__init__('Package for class "%s" '
                                                     'is not found' % name)


class NoObjectFoundError(Exception):
    def __init__(self, object_id):
        super(NoObjectFoundError, self).__init__(
            'Object "%s" is not found in object store' % object_id)


class MethodNotExposed(Exception):
    pass


class AmbiguousMethodName(Exception):
    def __init__(self, name):
        super(AmbiguousMethodName, self).__init__(
            'Found more than one method "%s"' % name)


class AmbiguousClassName(Exception):
    def __init__(self, name):
        super(AmbiguousClassName, self).__init__(
            'Found more than one version of class "%s"' % name)


class DslContractSyntaxError(Exception):
    pass


class ContractViolationException(Exception):
    def __init__(self, *args, **kwargs):
        super(ContractViolationException, self).__init__(*args, **kwargs)
        self._path = ''

    @property
    def path(self):
        return self._path

    @path.setter
    def path(self, value):
        self._path = value


class ValueIsMissingError(Exception):
    pass


class DslSyntaxError(Exception):
    pass


class PropertyAccessError(Exception):
    pass


class AmbiguousPropertyNameError(PropertyAccessError):
    def __init__(self, name):
        super(AmbiguousPropertyNameError, self).__init__(
            'Found more than one property "%s"' % name)


class NoWriteAccess(PropertyAccessError):
    def __init__(self, name):
        super(NoWriteAccess, self).__init__(
            'Property "%s" is immutable to the caller' % name)


class NoWriteAccessError(PropertyAccessError):
    def __init__(self, name):
        super(NoWriteAccessError, self).__init__(
            'Property "%s" is immutable to the caller' % name)


class PropertyReadError(PropertyAccessError):
    def __init__(self, name, murano_class):
        super(PropertyAccessError, self).__init__(
            'Property "%s" in class "%s" cannot be read' %
            (name, murano_class.name))


class PropertyWriteError(PropertyAccessError):
    def __init__(self, name, murano_class):
        super(PropertyAccessError, self).__init__(
            'Property "%s" in class "%s" cannot be written' %
            (name, murano_class.name))


class UninitializedPropertyAccessError(PropertyAccessError):
    def __init__(self, name, murano_class):
        super(PropertyAccessError, self).__init__(
            'Access to uninitialized property '
            '"%s" in class "%s" is forbidden' % (name, murano_class.name))


class CircularExpressionDependenciesError(Exception):
    pass


class InvalidLhsTargetError(Exception):
    def __init__(self, target):
        super(InvalidLhsTargetError, self).__init__(
            'Invalid assignment target "%s"' % target)


class InvalidInheritanceError(Exception):
    pass


class ObjectDestroyedError(Exception):
    def __init__(self, obj):
        super(ObjectDestroyedError, self).__init__(
            'Object {0} is already destroyed'.format(obj))



#    Copyright (c) 2014 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import collections


from murano.dsl import dsl_types


class AttributeStore(object):
    def __init__(self):
        self._attributes = collections.defaultdict(lambda: {})

    @staticmethod
    def _get_attribute_key(tagged_object, owner_type, name):
        if isinstance(owner_type, dsl_types.MuranoTypeReference):
            owner_type = owner_type.type
        if isinstance(tagged_object, dsl_types.MuranoObjectInterface):
            tagged_object = tagged_object.object
        return tagged_object.object_id, (owner_type.name, name)

    def get(self, tagged_object, owner_type, name):
        key1, key2 = self._get_attribute_key(tagged_object, owner_type, name)
        return self._attributes[key1].get(key2)

    def set(self, tagged_object, owner_type, name, value):
        if isinstance(value, dsl_types.MuranoObjectInterface):
            value = value.id
        elif isinstance(value, dsl_types.MuranoObject):
            value = value.object_id
        key1, key2 = self._get_attribute_key(tagged_object, owner_type, name)
        if value is None:
            self._attributes[key1].pop(key2, None)
        else:
            self._attributes[key1][key2] = value

    def serialize(self, known_objects):
        return [
            [key1, key2[0], key2[1], value]
            for key1, inner in self._attributes.items()
            for key2, value in inner.items()
            if key1 in known_objects
        ]

    def load(self, data):
        for item in data:
            if item[3] is not None:
                self._attributes[item[0]][(item[1], item[2])] = item[3]

    def forget_object(self, obj):
        if isinstance(obj, dsl_types.MuranoObjectInterface):
            obj = obj.id
        elif isinstance(obj, dsl_types.MuranoObject):
            obj = obj.object_id
        self._attributes.pop(obj, None)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""empty message

Revision ID: 006
Revises: None
Create Date: 2015-02-15 12:14:12

"""
from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = '006'
down_revision = '005'

MYSQL_ENGINE = 'InnoDB'
MYSQL_CHARSET = 'utf8'


def upgrade():
    op.add_column('task', sa.Column('result', sa.types.Text()))
    # end Alembic commands #


def downgrade():
    op.drop_column('task', 'result')
    # end Alembic commands #



#    Copyright (c) 2013 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from murano.common import utils
from murano.tests.unit import base


class TraverseHelperTests(base.MuranoTestCase):
    def test_root_get_with_dict(self):
        source = {'attr': True}
        value = utils.TraverseHelper.get('/', source)
        self.assertEqual(value, source)

    def test_root_get_with_list(self):
        source = [{'attr': True}]
        value = utils.TraverseHelper.get('/', source)
        self.assertListEqual(value, source)

    def test_root_get_with_value_type(self):
        source = 'source'
        value = utils.TraverseHelper.get('/', source)
        self.assertEqual(source, value)

    def test_attribute_get(self):
        source = {'attr': True}
        value = utils.TraverseHelper.get('/attr', source)
        self.assertTrue(value)

    def test_nested_attribute_get(self):
        source = {'obj': {'attr': True}}
        value = utils.TraverseHelper.get('/obj/attr', source)
        self.assertTrue(value)

    def test_list_item_attribute_get(self):
        source = {'obj': [
            {'?': {'id': '1'}, 'value': 1},
            {'?': {'id': '2s'}, 'value': 2},
        ]}
        value = utils.TraverseHelper.get('/obj/2s/value', source)
        self.assertEqual(2, value)

    def test_list_item_attribute_get_by_index(self):
        source = {'obj': [
            {'?': {'id': 'guid1'}, 'value': 1},
            {'?': {'id': 'guid2'}, 'value': 2}
        ]}
        value = utils.TraverseHelper.get('/obj/1/value', source)
        self.assertEqual(2, value)

    def test_attribute_set(self):
        source = {'attr': True}
        utils.TraverseHelper.update('/newAttr', False, source)
        value = utils.TraverseHelper.get('/newAttr', source)
        self.assertFalse(value)

    def test_attribute_update(self):
        source = {'attr': True}
        utils.TraverseHelper.update('/attr', False, source)
        value = utils.TraverseHelper.get('/attr', source)
        self.assertFalse(value)

    def test_nested_attribute_update(self):
        source = {'obj': {'attr': True}}
        utils.TraverseHelper.update('/obj/attr', False, source)
        value = utils.TraverseHelper.get('/obj/attr', source)
        self.assertFalse(value)

    def test_adding_item_to_list(self):
        source = {'attr': [1, 2, 3]}
        utils.TraverseHelper.insert('/attr', 4, source)
        value = utils.TraverseHelper.get('/attr', source)
        self.assertListEqual(value, [1, 2, 3, 4])

    def test_nested_adding_item_to_list(self):
        source = {'obj': {'attr': [1, 2, 3]}}
        utils.TraverseHelper.insert('/obj/attr', 4, source)
        value = utils.TraverseHelper.get('/obj/attr', source)
        self.assertListEqual(value, [1, 2, 3, 4])

    def test_extending_list_with_list(self):
        source = {'attr': [1, 2, 3]}
        utils.TraverseHelper.extend('/attr', [4, 5], source)
        value = utils.TraverseHelper.get('/attr', source)
        self.assertListEqual(value, [1, 2, 3, 4, 5])

    def test_nested_extending_list_with_list(self):
        source = {'obj': {'attr': [1, 2, 3]}}
        utils.TraverseHelper.extend('/obj/attr', [4, 5], source)
        value = utils.TraverseHelper.get('/obj/attr', source)
        self.assertListEqual(value, [1, 2, 3, 4, 5])

    def test_attribute_remove_from_dict(self):
        source = {'attr1': False, 'attr2': True}
        utils.TraverseHelper.remove('/attr1', source)
        value = utils.TraverseHelper.get('/', source)
        self.assertEqual(value, {'attr2': True})

    def test_nested_attribute_remove_from_dict(self):
        source = {'obj': {'attr1': False, 'attr2': True}}
        utils.TraverseHelper.remove('/obj/attr1', source)
        value = utils.TraverseHelper.get('/obj', source)
        self.assertEqual(value, {'attr2': True})

    def test_nested_attribute_remove_from_list_by_id(self):
        source = {'obj': [{'?': {'id': 'id1'}}, {'?': {'id': 'id2'}}]}
        utils.TraverseHelper.remove('/obj/id1', source)
        value = utils.TraverseHelper.get('/obj', source)
        self.assertListEqual(value, [{'?': {'id': 'id2'}}])

    def test_nested_attribute_remove_from_list_by_index(self):
        source = {'obj': [{'?': {'id': 'id1'}}, {'?': {'id': 'id2'}}]}
        utils.TraverseHelper.remove('/obj/0', source)
        value = utils.TraverseHelper.get('/obj', source)
        self.assertListEqual(value, [{'?': {'id': 'id2'}}])



#    Copyright (c) 2015 Mirantis, Inc.
#    Copyright 2017 AT&T Corporation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import json
from unittest import mock

from oslo_serialization import base64
from webob import response

from murano.cfapi import cfapi as api
from murano.common import wsgi
from murano.db import models
from murano.tests.unit import base

from muranoclient.common import exceptions


class TestController(base.MuranoTestCase):
    def setUp(self):
        super(TestController, self).setUp()
        self.controller = api.Controller()

        self.request = mock.MagicMock()
        auth = 'Basic {}'.format(base64.encode_as_text(b'test:test'))
        self.request.headers = {'Authorization': auth,
                                'X-Auth-Token': 'foo-bar',
                                'X-Project-Id': 'bar-baz'}
        self.addCleanup(mock.patch.stopall)

    @mock.patch('murano.common.policy.check_is_admin')
    @mock.patch('murano.cfapi.cfapi._get_muranoclient')
    def test_list(self, mock_client, mock_policy):

        pkg0 = mock.MagicMock()
        pkg0.id = 'xxx'
        pkg0.name = 'foo'
        pkg0.description = 'stub pkg'

        mock_client.return_value.packages.filter =\
            mock.MagicMock(return_value=[pkg0])

        answer = {'services': [{'bindable': True,
                                'description': pkg0.description,
                                'id': pkg0.id,
                                'name': pkg0.name,
                                'plans': [{'description': ('Default plan for '
                                                           'the service '
                                                           '{name}').format(
                                                               name=pkg0.name),
                                           'id': 'xxx-1',
                                           'name': 'default'}],
                                'tags': []}]}

        resp = self.controller.list(self.request)
        self.assertEqual(answer, resp)

    @mock.patch('murano.common.policy.check_is_admin')
    @mock.patch('murano.cfapi.cfapi._get_muranoclient')
    @mock.patch('murano.db.services.cf_connections.set_instance_for_service')
    @mock.patch('murano.db.services.cf_connections.get_environment_for_space')
    @mock.patch('murano.db.services.cf_connections.get_tenant_for_org')
    def test_provision_from_scratch(self, mock_get_tenant,
                                    mock_get_environment, mock_is, mock_client,
                                    mock_policy):

        body = {"space_guid": "s1-p1",
                "organization_guid": "o1-r1",
                "plan_id": "p1-l1",
                "service_id": "s1-e1",
                "parameters": {'some_parameter': 'value',
                               '?': {}}}
        self.request.body = json.dumps(body)

        mock_get_environment.return_value = '555-555'
        mock_client.return_value = mock.MagicMock()

        resp = self.controller.provision(self.request, {}, '111-111')

        self.assertIsInstance(resp, response.Response)

    @mock.patch('murano.common.policy.check_is_admin')
    @mock.patch('murano.cfapi.cfapi._get_muranoclient')
    @mock.patch('murano.db.services.cf_connections.set_instance_for_service')
    @mock.patch('murano.db.services.cf_connections.set_environment_for_space')
    @mock.patch('murano.db.services.cf_connections.set_tenant_for_org')
    @mock.patch('murano.db.services.cf_connections.get_environment_for_space')
    @mock.patch('murano.db.services.cf_connections.get_tenant_for_org')
    def test_provision_existent(self, mock_get_tenant,
                                mock_get_environment, mock_set_tenant,
                                mock_set_environment, mock_is, mock_client,
                                mock_policy):

        body = {"space_guid": "s1-p1",
                "organization_guid": "o1-r1",
                "plan_id": "p1-l1",
                "service_id": "s1-e1",
                "parameters": {'some_parameter': 'value',
                               '?': {}}}
        self.request.body = json.dumps(body)

        mock_package = mock_client.return_value.packages.get
        mock_package.return_value = mock.MagicMock()
        mock_get_environment.side_effect = AttributeError
        mock_get_tenant.side_effect = AttributeError

        resp = self.controller.provision(self.request, {}, '111-111')

        self.assertIsInstance(resp, response.Response)

    @mock.patch('murano.cfapi.cfapi._get_muranoclient')
    @mock.patch('murano.db.services.cf_connections.get_service_for_instance')
    def test_deprovision(self, mock_get_si, mock_client):
        service = mock.MagicMock()
        service.service_id = '111-111'
        service.tenant_id = '222-222'
        service.env_id = '333-333'
        mock_get_si.return_value = service

        resp = self.controller.deprovision(self.request, '555-555')

        self.assertIsInstance(resp, response.Response)

    @mock.patch('murano.cfapi.cfapi._get_muranoclient')
    @mock.patch('murano.db.services.cf_connections.get_service_for_instance')
    def test_bind(self, mock_get_si, mock_client):
        service = mock.MagicMock()
        service.service_id = '111-111'
        service.tenant_id = '222-222'
        service.env_id = '333-333'
        mock_get_si.return_value = service

        services = [{'id': 'xxx-xxx-xxx',
                     '?': {'id': '111-111',
                           '_actions': {
                               'dafsa_getCredentials': {
                                   'name': 'getCredentials'}}},
                     'instance': {}}]
        mock_client.return_value.environments.get =\
            mock.MagicMock(return_value=mock.MagicMock(services=services))

        mock_client.return_value.actions.get_result =\
            mock.MagicMock(return_value={'result': {'smthg': 'nothing'}})
        nice_resp = {'credentials': {'smthg': 'nothing'}}
        resp = self.controller.bind(self.request, {}, '555-555', '666-666')

        self.assertEqual(nice_resp, resp)

    def test_package_to_service(self):
        mock_package = mock.Mock(
            spec_set=models.Package, id='foo_package_id',
            description='a' * 257, tags=[mock.sentinel.package_tag])
        mock_package.configure_mock(name=mock.sentinel.package_name)

        expected_service = {
            'id': 'foo_package_id',
            'name': mock.sentinel.package_name,
            'description': 'a' * 253 + ' ...',
            'bindable': True,
            'tags': [mock.sentinel.package_tag],
            'plans': [{
                'id': 'foo_package_id-1',
                'name': 'default',
                'description':
                'Default plan for the service sentinel.package_name'
            }]
        }
        service = self.controller._package_to_service(mock_package)

        for key, val in expected_service.items():
            if key != 'plans':
                self.assertEqual(val, service[key])
            else:
                self.assertEqual(1, len(service['plans']))
                for plan_key, plan_val in expected_service['plans'][0].items():
                    self.assertEqual(plan_val, service['plans'][0][plan_key])

    def test_get_service_return_none(self):
        mock_env = mock.Mock(services=[])
        self.assertIsNone(self.controller._get_service(mock_env, None))

    @mock.patch.object(api, 'LOG', autospec=True)
    @mock.patch.object(api, 'db_cf', autospec=True)
    @mock.patch.object(api, '_get_muranoclient', autospec=True)
    def test_provision_except_http_not_found(self, mock_get_muranoclient,
                                             mock_db_cf, mock_log):
        test_body = {
            'space_guid': 'foo_space_guid',
            'organization_guid': 'foo_organization_guid',
            'plan_id': 'foo_plan_id',
            'service_id': 'foo_service_id',
            'parameters': {}
        }
        test_headers = {
            'X-Auth-Token': mock.sentinel.token
        }
        mock_request = mock.Mock(body=json.dumps(test_body),
                                 headers=test_headers)
        mock_muranoclient = mock.Mock()
        mock_muranoclient.environments.get.side_effect = \
            exceptions.HTTPNotFound
        mock_muranoclient.environments.create.return_value = \
            mock.Mock(id=mock.sentinel.alt_environment_id)
        mock_get_muranoclient.return_value = mock_muranoclient
        mock_db_cf.get_environment_for_space.return_value = \
            mock.sentinel.environment_id

        resp = self.controller.provision(mock_request, None, None)

        self.assertEqual(202, resp.status_code)
        self.assertEqual({}, resp.json_body)
        mock_db_cf.get_environment_for_space.assert_called_once_with(
            'foo_space_guid')
        mock_muranoclient.environments.get.assert_called_once_with(
            mock.sentinel.environment_id)
        mock_log.info.assert_has_calls([
            mock.call("Can not find environment_id sentinel.environment_id"
                      ", will create a new one."),
            mock.call("Cloud Foundry foo_space_guid remapped to "
                      "sentinel.alt_environment_id")
        ])

    @mock.patch.object(api, 'uuid', autospec=True)
    @mock.patch.object(api, 'db_cf', autospec=True)
    @mock.patch.object(api, '_get_muranoclient', autospec=True)
    def test_provision_with_dict_parameters(self, mock_get_muranoclient,
                                            mock_db_cf, mock_uuid):
        test_body = {
            'space_guid': 'foo_space_guid',
            'organization_guid': 'foo_organization_guid',
            'plan_id': 'foo_plan_id',
            'service_id': 'foo_service_id',
            'parameters': {
                'foo': {
                    '?': {
                        'bar': 'baz'
                    }
                }
            }
        }
        test_headers = {
            'X-Auth-Token': mock.sentinel.token
        }
        mock_request = mock.Mock(body=json.dumps(test_body),
                                 headers=test_headers)
        mock_package = mock_get_muranoclient.return_value.packages.get.\
            return_value
        mock_service = mock.MagicMock()
        self.controller._make_service = mock.Mock(return_value=mock_service)
        mock_uuid.uuid4.return_value.hex = mock.sentinel.uuid

        resp = self.controller.provision(mock_request, None, None)

        self.assertEqual(202, resp.status_code)
        self.assertEqual({}, resp.json_body)
        self.controller._make_service.assert_called_once_with(
            'foo_space_guid', mock_package, 'foo_plan_id')
        mock_service.update.assert_called_once_with(
            {'foo': {'?': {'bar': 'baz', 'id': mock.sentinel.uuid}}})

    @mock.patch.object(api, 'db_cf', autospec=True)
    def test_deprovision_with_missing_service(self, mock_db_cf):
        mock_db_cf.get_service_for_instance.return_value = None
        self.assertEqual({}, self.controller.deprovision(None, None))
        mock_db_cf.get_service_for_instance.assert_called_once_with(None)

    @mock.patch.object(api, 'db_cf', autospec=True)
    def test_bind_with_missing_service(self, mock_db_cf):
        mock_db_cf.get_service_for_instance.return_value = None
        self.assertEqual({}, self.controller.bind(None, None, None, None))
        mock_db_cf.get_service_for_instance.assert_called_once_with(None)

    @mock.patch.object(api, 'LOG', autospec=True)
    @mock.patch.object(api.Controller, '_get_service', autospec=True)
    @mock.patch.object(api, '_get_muranoclient', autospec=True)
    @mock.patch.object(api, 'db_cf', autospec=True)
    def test_bind_except_key_error(self, mock_db_cf, mock_get_muranoclient,
                                   mock_get_service, mock_log):
        mock_db_service = mock.Mock(
            service_id=mock.sentinel.service_id,
            environment_id=mock.sentinel.environment_id)
        test_service = {}
        test_request = mock.Mock(
            headers={'X-Auth-Token': mock.sentinel.auth_token})

        mock_db_cf.get_service_for_instance.return_value = mock_db_service
        mock_get_service.return_value = test_service

        m_cli = mock_get_muranoclient.return_value
        m_env = m_cli.environments.get.return_value

        resp = self.controller.bind(test_request, None, None, None)

        self.assertEqual(500, resp.status_code)
        m_cli.environments.get.assert_called_once_with(
            mock.sentinel.environment_id, mock.ANY)
        mock_get_service.assert_called_once_with(self.controller, m_env,
                                                 mock.sentinel.service_id)
        mock_log.warning.assert_called_once_with(
            "This application doesn't have actions at all")

    @mock.patch.object(api, 'LOG', autospec=True)
    @mock.patch.object(api.Controller, '_get_service', autospec=True)
    @mock.patch.object(api, '_get_muranoclient', autospec=True)
    @mock.patch.object(api, 'db_cf', autospec=True)
    def test_bind_without_get_credentials(self, mock_db_cf,
                                          mock_get_muranoclient,
                                          mock_get_service, mock_log):
        mock_db_service = mock.Mock(
            service_id=mock.sentinel.service_id,
            environment_id=mock.sentinel.environment_id)
        test_service = {'?': {'_actions': []}}
        test_request = mock.Mock(
            headers={'X-Auth-Token': mock.sentinel.auth_token})

        mock_db_cf.get_service_for_instance.return_value = mock_db_service
        mock_get_service.return_value = test_service

        m_cli = mock_get_muranoclient.return_value
        m_env = m_cli.environments.get.return_value

        resp = self.controller.bind(test_request, None, None, None)

        self.assertEqual(500, resp.status_code)
        m_cli.environments.get.assert_called_once_with(
            mock.sentinel.environment_id, mock.ANY)
        mock_get_service.assert_called_once_with(self.controller, m_env,
                                                 mock.sentinel.service_id)
        mock_log.warning.assert_called_once_with(
            "This application doesn't have action getCredentials")

    def test_unbind(self):
        self.assertEqual({}, self.controller.unbind(None, None, None))

    @mock.patch.object(api, 'LOG', autospec=True)
    @mock.patch.object(api, 'db_cf', autospec=True)
    def test_get_last_operation_without_service(self, mock_db_cf, mock_log):
        mock_db_cf.get_service_for_instance.return_value = None

        resp = self.controller.get_last_operation(
            None, mock.sentinel.instance_id)

        self.assertEqual(410, resp.status_code)
        self.assertEqual({}, resp.json_body)
        mock_log.warning.assert_called_once_with(
            'Requested service for instance sentinel.instance_id is not '
            'found')

    @mock.patch.object(api, '_get_muranoclient', autospec=True)
    @mock.patch.object(api, 'db_cf', autospec=True)
    def test_get_last_operation_succeeded(self, mock_db_cf,
                                          mock_get_muranoclient):
        mock_service = mock.Mock(
            environment_id=mock.sentinel.environment_id)
        mock_request = mock.Mock(
            headers={'X-Auth-Token': mock.sentinel.auth_token})
        mock_db_cf.get_service_for_instance.return_value = mock_service

        m_cli = mock_get_muranoclient.return_value
        m_env = m_cli.environments.get.return_value
        m_env.status = 'ready'

        resp = self.controller.get_last_operation(
            mock_request, mock.sentinel.instance_id)

        self.assertEqual(200, resp.status_code)
        self.assertEqual(
            {'state': 'succeeded', 'description': 'operation succeed'},
            resp.json_body)
        mock_get_muranoclient.assert_called_once_with(mock.sentinel.auth_token,
                                                      mock_request)
        m_cli.environments.get.assert_called_once_with(
            mock.sentinel.environment_id)

    @mock.patch.object(api, '_get_muranoclient', autospec=True)
    @mock.patch.object(api, 'db_cf', autospec=True)
    def test_get_last_operation_in_progress(self, mock_db_cf,
                                            mock_get_muranoclient):
        mock_service = mock.Mock(
            environment_id=mock.sentinel.environment_id)
        mock_request = mock.Mock(
            headers={'X-Auth-Token': mock.sentinel.auth_token})
        mock_db_cf.get_service_for_instance.return_value = mock_service

        m_cli = mock_get_muranoclient.return_value
        m_env = m_cli.environments.get.return_value

        progress_statuses = ['pending', 'deleting', 'deploying']
        for status in progress_statuses:
            m_env.status = status

            resp = self.controller.get_last_operation(
                mock_request, mock.sentinel.instance_id)

            self.assertEqual(202, resp.status_code)
            self.assertEqual(
                {'state': 'in progress',
                 'description': 'operation in progress'},
                resp.json_body)
            mock_get_muranoclient.assert_called_once_with(
                mock.sentinel.auth_token, mock_request)
            m_cli.environments.get.assert_called_once_with(
                mock.sentinel.environment_id)
            mock_get_muranoclient.reset_mock()
            m_cli.environments.get.reset_mock()

    @mock.patch.object(api, '_get_muranoclient', autospec=True)
    @mock.patch.object(api, 'db_cf', autospec=True)
    def test_get_last_operation_failed(self, mock_db_cf,
                                       mock_get_muranoclient):
        mock_service = mock.Mock(
            environment_id=mock.sentinel.environment_id)
        mock_request = mock.Mock(
            headers={'X-Auth-Token': mock.sentinel.auth_token})
        mock_db_cf.get_service_for_instance.return_value = mock_service

        m_cli = mock_get_muranoclient.return_value
        m_env = m_cli.environments.get.return_value

        failed_statuses = ['deploy failure', 'delete failure']
        for status in failed_statuses:
            m_env.status = status

            resp = self.controller.get_last_operation(
                mock_request, mock.sentinel.instance_id)

            self.assertEqual(200, resp.status_code)
            self.assertEqual(
                {'state': 'failed',
                 'description': '{0}. Please correct it manually'
                 .format(status)},
                resp.json_body)
            mock_get_muranoclient.assert_called_once_with(
                mock.sentinel.auth_token, mock_request)
            m_cli.environments.get.assert_called_once_with(
                mock.sentinel.environment_id)
            mock_get_muranoclient.reset_mock()
            m_cli.environments.get.reset_mock()

    @mock.patch.object(api, '_get_muranoclient', autospec=True)
    @mock.patch.object(api, 'db_cf', autospec=True)
    def test_get_last_operation_unknown_status(self, mock_db_cf,
                                               mock_get_muranoclient):
        mock_service = mock.Mock(
            environment_id=mock.sentinel.environment_id)
        mock_request = mock.Mock(
            headers={'X-Auth-Token': mock.sentinel.auth_token})
        mock_db_cf.get_service_for_instance.return_value = mock_service

        m_cli = mock_get_muranoclient.return_value
        m_env = m_cli.environments.get.return_value

        m_env.status = 'unknown'

        resp = self.controller.get_last_operation(
            mock_request, mock.sentinel.instance_id)

        self.assertEqual(500, resp.status_code)
        self.assertEqual(
            {'state': 'unknown', 'description': 'operation unknown'},
            resp.json_body)
        mock_get_muranoclient.assert_called_once_with(
            mock.sentinel.auth_token, mock_request)
        m_cli.environments.get.assert_called_once_with(
            mock.sentinel.environment_id)

    @mock.patch.object(api, 'muranoclient', autospec=True)
    @mock.patch.object(api, 'glare_client', autospec=True)
    def test_get_muranoclient(self, mock_glare_client, mock_muranoclient):
        self._override_conf()

        m_artifacts_client = mock.Mock()
        m_muranoclient = mock.Mock()
        mock_glare_client.Client.return_value = m_artifacts_client
        mock_muranoclient.Client.return_value = m_muranoclient

        client = api._get_muranoclient(mock.sentinel.token_id, None)

        self.assertEqual(m_muranoclient, client)
        mock_glare_client.Client.assert_called_once_with(
            endpoint='foo_glare_url', token=mock.sentinel.token_id,
            insecure=True, key_file='foo_key_file', ca_file='foo_ca_file',
            cert_file='foo_cert_file', type_name='murano', type_version=1)
        mock_muranoclient.Client.assert_called_once_with(
            1, 'foo_murano_url', token=mock.sentinel.token_id,
            artifacts_client=m_artifacts_client)

    @mock.patch.object(api, 'LOG', autospec=True)
    @mock.patch.object(api, 'muranoclient', autospec=True)
    @mock.patch.object(api, 'glare_client', autospec=True)
    def test_get_muranoclient_without_urls(self, mock_glare_client,
                                           mock_muranoclient, mock_log):
        self._override_conf(without_urls=True)

        m_artifacts_client = mock.Mock()
        m_muranoclient = mock.Mock()
        mock_glare_client.Client.return_value = m_artifacts_client
        mock_muranoclient.Client.return_value = m_muranoclient
        mock_request = mock.Mock(endpoints={'murano': None})

        client = api._get_muranoclient(mock.sentinel.token_id, mock_request)

        self.assertEqual(m_muranoclient, client)
        mock_glare_client.Client.assert_called_once_with(
            endpoint=None, token=mock.sentinel.token_id,
            insecure=True, key_file='foo_key_file', ca_file='foo_ca_file',
            cert_file='foo_cert_file', type_name='murano', type_version=1)
        mock_muranoclient.Client.assert_called_once_with(
            1, None, token=mock.sentinel.token_id,
            artifacts_client=m_artifacts_client)
        mock_log.error.assert_has_calls([
            mock.call('No glare url is specified and no "artifact" '
                      'service is registered in keystone.'),
            mock.call('No murano url is specified and no '
                      '"application-catalog" service is registered in '
                      'keystone.')
        ])

    def _override_conf(self, without_urls=False):
        if without_urls:
            self.override_config('url', None, group='glare')
            self.override_config('url', None, group='murano')
        else:
            self.override_config('url', 'foo_glare_url', group='glare')
            self.override_config('url', 'foo_murano_url', group='murano')
        for arg, group, override_val in (
                ('packages_service', 'engine', 'glare'),
                ('insecure', 'glare', True),
                ('key_file', 'glare', 'foo_key_file'),
                ('ca_file', 'glare', 'foo_ca_file'),
                ('cert_file', 'glare', 'foo_cert_file')):
            self.override_config(arg, override_val, group=group)

    def test_resource(self):
        self.assertIsInstance(api.create_resource(), wsgi.Resource)



#    Copyright (c) 2014 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
import itertools
from keystoneauth1 import loading as ks_loading

import oslo_service.sslutils

import murano.common.cf_config
import murano.common.config
import murano.common.wsgi

__all__ = [
    'list_opts',
    'list_cfapi_opts',
]


def build_list(opt_list):
    return list(itertools.chain(*opt_list))


# List of *all* options in [DEFAULT] namespace of murano.
# Any new option list or option needs to be registered here.
_opt_lists = [
    ('engine', murano.common.config.engine_opts),
    ('rabbitmq', murano.common.config.rabbit_opts),
    ('heat',
     murano.common.config.heat_opts +
     ks_loading.get_session_conf_options()),
    ('neutron',
     murano.common.config.neutron_opts +
     ks_loading.get_session_conf_options()),
    ('murano', murano.common.config.murano_opts +
     ks_loading.get_session_conf_options()),
    ('glare',
     murano.common.config.glare_opts +
     ks_loading.get_session_conf_options()),
    ('mistral',
     murano.common.config.mistral_opts +
     ks_loading.get_session_conf_options()),
    ('networking', murano.common.config.networking_opts),
    ('stats', murano.common.config.stats_opts),
    ('murano_auth',
     murano.common.config.murano_auth_opts +
     ks_loading.get_session_conf_options() +
     ks_loading.get_auth_common_conf_options() +
     ks_loading.get_auth_plugin_conf_options('password') +
     ks_loading.get_auth_plugin_conf_options('v2password') +
     ks_loading.get_auth_plugin_conf_options('v3password')),
    (None, build_list([
        murano.common.config.metadata_dir,
        murano.common.config.bind_opts,
        murano.common.config.file_server,
        murano.common.wsgi.wsgi_opts,
    ])),
]

_cfapi_opt_lists = [
    ('cfapi', murano.common.cf_config.cfapi_opts),
    ('glare',
     murano.common.config.glare_opts +
     ks_loading.get_session_conf_options())
]

_opt_lists.extend(oslo_service.sslutils.list_opts())


def list_opts():
    """Return a list of oslo.config options available in Murano.

    Each element of the list is a tuple. The first element is the name of the
    group under which the list of elements in the second element will be
    registered. A group name of None corresponds to the [DEFAULT] group in
    config files.

    This function is also discoverable via the 'murano' entry point
    under the 'oslo.config.opts' namespace.

    The purpose of this is to allow tools like the Oslo sample config file
    generator to discover the options exposed to users by Murano.

    :returns: a list of (group_name, opts) tuples
    """
    return [(g, copy.deepcopy(o)) for g, o in _opt_lists]


def list_cfapi_opts():
    """Return a list of oslo_config options available in service broker."""
    return [(g, copy.deepcopy(o)) for g, o in _cfapi_opt_lists]



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json as jsonlib

import yaml as yamllib
from yaql.language import specs
from yaql.language import yaqltypes

from murano.dsl import constants
from murano.dsl import dsl
from murano.dsl import dsl_types
from murano.dsl import helpers

if hasattr(yamllib, 'CSafeLoader'):
    yaml_loader = yamllib.CSafeLoader
else:
    yaml_loader = yamllib.SafeLoader


def _construct_yaml_str(self, node):
    # Override the default string handling function
    # to always return unicode objects
    return self.construct_scalar(node)


yaml_loader.add_constructor(u'tag:yaml.org,2002:str', _construct_yaml_str)
# Unquoted dates like 2013-05-23 in yaml files get loaded as objects of type
# datetime.data which causes problems in API layer when being processed by
# oslo.serialization.jsonutils. Therefore, make unicode string out of
# timestamps until jsonutils can handle dates.
yaml_loader.add_constructor(u'tag:yaml.org,2002:timestamp',
                            _construct_yaml_str)


@dsl.name('io.murano.system.Resources')
class ResourceManager(object):
    def __init__(self, context):
        murano_class = helpers.get_type(helpers.get_caller_context(context))
        self._package = murano_class.package

    @staticmethod
    @specs.parameter('owner', dsl.MuranoTypeParameter(nullable=True))
    @specs.inject('receiver', yaqltypes.Receiver())
    @specs.meta(constants.META_NO_TRACE, True)
    def string(receiver, name, owner=None, binary=False):
        path = ResourceManager._get_package(owner, receiver).get_resource(name)
        mode = 'rb' if binary else 'rU'
        with open(path, mode) as file:
            return file.read()

    @classmethod
    @specs.parameter('owner', dsl.MuranoTypeParameter(nullable=True))
    @specs.inject('receiver', yaqltypes.Receiver())
    @specs.meta(constants.META_NO_TRACE, True)
    def json(cls, receiver, name, owner=None):
        return jsonlib.loads(cls.string(receiver, name, owner))

    @classmethod
    @specs.parameter('owner', dsl.MuranoTypeParameter(nullable=True))
    @specs.inject('receiver', yaqltypes.Receiver())
    @specs.meta(constants.META_NO_TRACE, True)
    def yaml(cls, receiver, name, owner=None):
        # NOTE(kzaitsev, Sam Pilla) Bandit will raise an issue here,
        # because it thinks that we're using an unsafe yaml.load.
        # However we're passing a SafeLoader here
        # (see definition of `yaml_loader` in this file; L27-30)
        # so a `nosec` was added to ignore the false positive report.
        return yamllib.load(  # nosec
            cls.string(receiver, name, owner), Loader=yaml_loader)

    @staticmethod
    def _get_package(owner, receiver):
        if owner is None:
            if isinstance(receiver, dsl_types.MuranoObjectInterface):
                return receiver.extension._package
            murano_class = helpers.get_type(helpers.get_caller_context())
        else:
            murano_class = owner.type
        return murano_class.package



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Add the is_public column to the environment-template for public
environment template functionality.

Revision ID: 011
Revises: table template

"""

from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = '011'
down_revision = '010'

MYSQL_ENGINE = 'InnoDB'
MYSQL_CHARSET = 'utf8'


def upgrade():
    op.add_column('environment-template',
                  sa.Column('is_public', sa.Boolean(),
                            default=False, nullable=True))
    # end Alembic commands #


def downgrade():
    op.drop_column('environment-template', 'is_public')
    # end Alembic commands #



# Copyright 2016 AT&T Corp
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from oslo_db import exception
from oslo_utils import timeutils

from murano.db.services import instances
from murano.tests.unit import base as test_base


@mock.patch('murano.db.services.instances.db_session')
class TestInstances(test_base.MuranoTestCase):

    @mock.patch('murano.db.services.instances.timeutils')
    @mock.patch('murano.db.services.instances.models')
    def test_track_instance(self, mock_models, mock_timeutils,
                            mock_db_session):
        mock_env = mock.MagicMock(tenant_id='test_tenant_id')
        mock_db_session.get_session().query().get.return_value = mock_env
        now = timeutils.utcnow_ts()
        mock_timeutils.utcnow_ts.return_value = now

        track_instance = instances.InstanceStatsServices.track_instance
        track_instance('test_instance_id', 'test_env_id', 'test_type',
                       'test_type_name', 'test_type_title', unit_count=1)

        mock_models.Instance.assert_called_once_with()
        self.assertEqual('test_instance_id',
                         mock_models.Instance().instance_id)
        self.assertEqual('test_env_id', mock_models.Instance().environment_id)
        self.assertEqual('test_instance_id',
                         mock_models.Instance().instance_id)
        self.assertEqual('test_tenant_id', mock_models.Instance().tenant_id)
        self.assertEqual('test_type', mock_models.Instance().instance_type)
        self.assertEqual(now, mock_models.Instance().created)
        self.assertIsNone(mock_models.Instance().destroyed)
        self.assertEqual('test_type_name', mock_models.Instance().type_name)
        self.assertEqual('test_type_title', mock_models.Instance().type_title)
        self.assertEqual(1, mock_models.Instance().unit_count)
        mock_db_session.get_session().add.assert_called_once_with(
            mock_models.Instance())

    @mock.patch('murano.db.services.instances.models')
    def test_track_instance_except_duplicate_entry(self, _, mock_db_session):
        mock_db_session.get_session().add.side_effect =\
            exception.DBDuplicateEntry

        track_instance = instances.InstanceStatsServices.track_instance
        track_instance('test_instance_id', 'test_env_id', 'test_type',
                       'test_type_name', 'test_type_title')

        self.assertEqual(1, mock_db_session.get_session().execute.call_count)

    @mock.patch('murano.db.services.instances.timeutils')
    def test_destroy_instance(self, mock_timeutils, mock_db_session):
        mock_instance = mock.MagicMock(destroyed=False)
        mock_db_session.get_session().query().get.return_value = mock_instance
        now = timeutils.utcnow_ts()
        mock_timeutils.utcnow_ts.return_value = now

        destroy_instance = instances.InstanceStatsServices.destroy_instance
        destroy_instance('test_instance_id', 'test_environment_id')

        self.assertEqual(now, mock_instance.destroyed)
        mock_instance.save.assert_called_once_with(
            mock_db_session.get_session())

    def test_get_aggregated_stats(self, mock_db_session):
        mock_db_session.get_session().query().filter().group_by().all\
            .return_value = [['1', '2', '3'], ['4', '5', '6']]

        get_stats = instances.InstanceStatsServices.get_aggregated_stats
        result = get_stats('test_env_id')

        expected_result = [
            {'type': 1, 'duration': 2, 'count': 3},
            {'type': 4, 'duration': 5, 'count': 6},
        ]

        self.assertEqual(expected_result, result)

    @mock.patch('murano.db.services.instances.timeutils')
    def test_get_raw_environment_stats(self, mock_timeutils, mock_db_session):
        now = timeutils.utcnow_ts()
        mock_timeutils.utcnow_ts.return_value = now
        mock_db_session.get_session().query().filter().filter().all.\
            return_value = [
                mock.MagicMock(instance_type='test_instance_type',
                               created=now - 1000,
                               destroyed=now,
                               type_name='test_type_name',
                               unit_count=1,
                               instance_id='test_instance_id',
                               type_title='test_type_title'),
                mock.MagicMock(instance_type='test_instance_type_2',
                               created=now - 2000,
                               destroyed=None,
                               type_name='test_type_name_2',
                               unit_count=2,
                               instance_id='test_instance_id_2',
                               type_title='test_type_title_2')
            ]
        mock_db_session.reset_mock()

        get_env_stats = instances.InstanceStatsServices.\
            get_raw_environment_stats
        result = get_env_stats('test_env_id', 'test_instance_id')

        expected_result = [
            {
                'type': 'test_instance_type',
                'duration': 1000,  # now - (now - 1000) = 1000
                'type_name': 'test_type_name',
                'unit_count': 1,
                'instance_id': 'test_instance_id',
                'type_title': 'test_type_title',
                'active': False
            },
            {
                'type': 'test_instance_type_2',
                'duration': 2000,  # now - (now - 2000) = 2000
                'type_name': 'test_type_name_2',
                'unit_count': 2,
                'instance_id': 'test_instance_id_2',
                'type_title': 'test_type_title_2',
                'active': True
            },
        ]

        self.assertEqual(expected_result, result)
        # Assert that two filters were performed, the second one for
        # the instance_id argument.
        self.assertEqual(1,
                         mock_db_session.get_session().query().filter.
                         call_count)
        self.assertEqual(1,
                         mock_db_session.get_session().query().filter().
                         filter.call_count)



#    Copyright (c) 2016 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import semantic_version
from yaql.language import specs
from yaql import yaqlization

from murano.dsl import dsl
from murano.dsl import dsl_types
from murano.dsl import helpers
from murano.dsl import meta


@specs.yaql_property(dsl_types.MuranoType)
@specs.name('name')
def type_name(murano_type):
    return murano_type.name


@specs.yaql_property(dsl_types.MuranoType)
@specs.name('usage')
def type_usage(murano_type):
    return murano_type.usage


@specs.yaql_property(dsl_types.MuranoClass)
def methods(murano_class):
    all_method_names = murano_class.all_method_names
    return tuple(
        murano_method
        for name in all_method_names
        if not name.startswith('__') and not name.startswith('.')
        for murano_method in murano_class.find_method(name)
    )


@specs.yaql_property(dsl_types.MuranoClass)
def properties(murano_class):
    all_property_names = murano_class.all_property_names
    return tuple(
        prop
        for prop_name in all_property_names
        if not prop_name.startswith('__') and not prop_name.startswith('.')
        for prop in murano_class.find_property(prop_name)
    )


@specs.yaql_property(dsl_types.MuranoClass)
def ancestors(murano_class):
    return tuple(murano_class.ancestors())


@specs.yaql_property(dsl_types.MuranoType)
def package(murano_type):
    return murano_type.package


@specs.yaql_property(dsl_types.MuranoClass)
@specs.name('version')
def type_version(murano_type):
    return murano_type.version


@specs.yaql_property(dsl_types.MuranoProperty)
@specs.name('name')
def property_name(murano_property):
    return murano_property.name


# TODO(ativelkov): add 'default' to return some wrapped YAQL expression
# @specs.yaql_property(dsl_types.MuranoProperty)
# @specs.name('default')
# def property_default(murano_property):
#     return murano_property.default


@specs.yaql_property(dsl_types.MuranoProperty)
@specs.name('has_default')
def property_has_default(murano_property):
    return murano_property.has_default


@specs.yaql_property(dsl_types.MuranoProperty)
@specs.name('usage')
def property_usage(murano_property):
    return murano_property.usage


@specs.yaql_property(dsl_types.MuranoProperty)
@specs.name('declaring_type')
def property_owner(murano_property):
    return murano_property.declaring_type


@specs.name('get_value')
@specs.parameter('property_', dsl_types.MuranoProperty)
@specs.parameter('object_', dsl.MuranoObjectParameter(
    nullable=True, decorate=False))
@specs.method
def property_get_value(context, property_, object_):
    if object_ is None:
        return helpers.get_executor().get_static_property(
            property_.declaring_type, name=property_.name, context=context)
    return object_.cast(property_.declaring_type).get_property(
        name=property_.name, context=context)


@specs.name('set_value')
@specs.parameter('property_', dsl_types.MuranoProperty)
@specs.parameter('object_', dsl.MuranoObjectParameter(
    nullable=True, decorate=False))
@specs.method
def property_set_value(context, property_, object_, value):
    if object_ is None:
        helpers.get_executor().set_static_property(
            property_.declaring_type,
            name=property_.name, value=value, context=context)
    else:
        object_.cast(property_.declaring_type).set_property(
            name=property_.name, value=value, context=context)


@specs.yaql_property(dsl_types.MuranoMethod)
@specs.name('name')
def method_name(murano_method):
    return murano_method.name


@specs.yaql_property(dsl_types.MuranoMethod)
def arguments(murano_method):
    if murano_method.arguments_scheme is None:
        return None
    return tuple(murano_method.arguments_scheme.values())


@specs.yaql_property(dsl_types.MuranoMethod)
@specs.name('declaring_type')
def method_owner(murano_method):
    return murano_method.declaring_type


@specs.parameter('method', dsl_types.MuranoMethod)
@specs.parameter('__object', dsl.MuranoObjectParameter(nullable=True))
@specs.name('invoke')
@specs.method
def method_invoke(context, method, __object, *args, **kwargs):
    return method.invoke(__object, args, kwargs, context)


@specs.yaql_property(dsl_types.MuranoPackage)
def types(murano_package):
    return tuple(
        murano_package.find_class(cls, False)
        for cls in murano_package.classes
    )


@specs.yaql_property(dsl_types.MuranoPackage)
@specs.name('name')
def package_name(murano_package):
    return murano_package.name


@specs.yaql_property(dsl_types.MuranoPackage)
@specs.name('version')
def package_version(murano_package):
    return murano_package.version


@specs.yaql_property(dsl_types.MuranoMethodArgument)
@specs.name('name')
def argument_name(method_argument):
    return method_argument.name

# TODO(ativelkov): add 'default' to return some wrapped YAQL expression
# @specs.yaql_property(dsl_types.MuranoMethodArgument)
# @specs.name('default')
# def argument_default(method_argument):
#     return method_argument.default


@specs.yaql_property(dsl_types.MuranoMethodArgument)
@specs.name('has_default')
def argument_has_default(method_argument):
    return method_argument.has_default


@specs.yaql_property(dsl_types.MuranoMethodArgument)
@specs.name('usage')
def argument_usage(method_argument):
    return method_argument.usage


@specs.yaql_property(dsl_types.MuranoMethodArgument)
@specs.name('declaring_method')
def argument_owner(method_argument):
    return method_argument.murano_method


@specs.yaql_property(dsl_types.MuranoType)
@specs.name('type')
def type_to_type_ref(murano_type):
    return murano_type.get_reference()


@specs.parameter('provider', meta.MetaProvider)
@specs.name('#property#meta')
def get_meta(context, provider):
    return provider.get_meta(context)


@specs.yaql_property(dsl_types.MuranoMetaClass)
def cardinality(murano_meta_class):
    return murano_meta_class.cardinality


@specs.yaql_property(dsl_types.MuranoMetaClass)
def targets(murano_meta_class):
    return murano_meta_class.targets


@specs.yaql_property(dsl_types.MuranoMetaClass)
def inherited(murano_meta_class):
    return murano_meta_class.inherited


def register(context):
    funcs = (
        type_name, type_usage, type_version, type_to_type_ref,
        methods, properties, ancestors, package,
        property_name, property_has_default, property_owner,
        property_usage, property_get_value, property_set_value,
        method_name, arguments, method_owner, method_invoke,
        types, package_name, package_version,
        argument_name, argument_has_default, argument_owner,
        argument_usage,
        cardinality, targets, inherited,
        get_meta
    )
    for f in funcs:
        context.register_function(f)


yaqlization.yaqlize(semantic_version.Version, whitelist=[
    'major', 'minor', 'patch', 'prerelease', 'build'
])



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
import shutil
import sys
import yaml
import zipfile

from murano.packages import exceptions
from murano.packages import package_base
from toscaparser.common import exception as csar_exception
from toscaparser.prereq import csar
from toscaparser.tosca_template import ToscaTemplate
from translator.hot.tosca_translator import TOSCATranslator

CSAR_RESOURCES_DIR_NAME = 'Resources/'
CSAR_FILES_DIR_NAME = 'CSARFiles/'
CSAR_ENV_DIR_NAME = 'CSAREnvironments/'


class YAQL(object):
    def __init__(self, expr):
        self.expr = expr


class Dumper(yaml.SafeDumper):
    pass


def yaql_representer(dumper, data):
    return dumper.represent_scalar(u'!yaql', data.expr)


Dumper.add_representer(YAQL, yaql_representer)


class CSARPackage(package_base.PackageBase):
    def __init__(self, format_name, runtime_version, source_directory,
                 manifest):
        super(CSARPackage, self).__init__(
            format_name, runtime_version, source_directory, manifest)

        self._translated_class = None
        self._source_directory = source_directory
        self._translated_ui = None

    @property
    def classes(self):
        return self.full_name,

    @property
    def requirements(self):
        return {}

    @property
    def ui(self):
        if not self._translated_ui:
            self._translated_ui = self._translate_ui()
        return self._translated_ui

    def get_class(self, name):
        if name != self.full_name:
            raise exceptions.PackageClassLoadError(
                name, 'Class not defined in this package')
        if not self._translated_class:
            self._translate_class()
        return self._translated_class, '<generated code>'

    def _translate_class(self):
        csar_file = os.path.join(self._source_directory, 'csar.zip')
        shutil.copy(csar_file, self.get_resource(self.full_name))

        if not os.path.isfile(csar_file):
            raise exceptions.PackageClassLoadError(
                self.full_name, 'File with class definition not found')

        csar_obj = csar.CSAR(csar_file)
        try:
            csar_obj.validate()
        except csar_exception.ValidationError as ve:
            raise exceptions.PackageFormatError('Not a CSAR archive: ' +
                                                str(ve))

        translated = {
            'Name': self.full_name,
            'Extends': 'io.murano.Application'
        }

        csar_envs_path = os.path.join(self._source_directory,
                                      CSAR_RESOURCES_DIR_NAME,
                                      CSAR_ENV_DIR_NAME)

        validate_csar_parameters = (not os.path.isdir(csar_envs_path) or
                                    not os.listdir(csar_envs_path))

        tosca = csar_obj.get_main_template_yaml()
        parameters = CSARPackage._build_properties(tosca,
                                                   validate_csar_parameters)
        parameters.update(CSARPackage._translate_outputs(tosca))
        translated['Properties'] = parameters
        hot = yaml.load(self._translate('tosca', csar_obj.csar,
                                        parameters, True))
        files = CSARPackage._translate_files(self._source_directory)

        template_file = os.path.join(self._source_directory,
                                     CSAR_RESOURCES_DIR_NAME, 'template.yaml')
        with open(template_file, 'w') as outfile:
            outfile.write(yaml.safe_dump(hot))
        translated.update(CSARPackage._generate_workflow(hot, files))
        self._translated_class = yaml.dump(translated, Dumper=Dumper,
                                           default_style='"')

    def _translate(self, sourcetype, path, parsed_params, a_file):
        output = None
        if sourcetype == "tosca":
            tosca = ToscaTemplate(path, parsed_params, a_file)
            translator = TOSCATranslator(tosca, parsed_params)
            output = translator.translate()
        return output

    @staticmethod
    def _build_properties(csar, csar_parameters):
        result = {
            'generatedHeatStackName': {
                'Contract': YAQL('$.string()'),
                'Usage': 'Out'
            },
            'hotEnvironment': {
                'Contract': YAQL('$.string()'),
                'Usage': 'In'
            }
        }

        if csar_parameters:
            params_dict = {}
            for key, value in (csar.get('parameters') or {}).items():
                param_contract = \
                    CSARPackage._translate_param_to_contract(value)
                params_dict[key] = param_contract
            result['templateParameters'] = {
                'Contract': params_dict,
                'Default': {},
                'Usage': 'In'
            }
        else:
            result['templateParameters'] = {
                'Contract': {},
                'Default': {},
                'Usage': 'In'
            }

        return result

    @staticmethod
    def _translate_param_to_contract(value):
        contract = '$'

        parameter_type = value['type']
        if parameter_type in ('string', 'comma_delimited_list', 'json'):
            contract += '.string()'
        elif parameter_type == 'integer':
            contract += '.int()'
        elif parameter_type == 'boolean':
            contract += '.bool()'
        else:
            raise ValueError('Unsupported parameter type ' + parameter_type)

        constraints = value.get('constraints') or []
        for constraint in constraints:
            translated = CSARPackage._translate_constraint(constraint)
            if translated:
                contract += translated

        result = YAQL(contract)
        return result

    @staticmethod
    def _translate_outputs(csar):
        result = {}
        for key in (csar.get('outputs') or {}).keys():
            result[key] = {
                "Contract": YAQL("$.string()"),
                "Usage": "Out"
            }
        return result

    @staticmethod
    def _translate_files(source_directory):
        source = os.path.join(source_directory, 'csar.zip')
        dest_dir = os.path.join(source_directory, CSAR_RESOURCES_DIR_NAME,
                                CSAR_FILES_DIR_NAME)
        with zipfile.ZipFile(source, "r") as z:
            z.extractall(dest_dir)
        csar_files_path = os.path.join(source_directory,
                                       CSAR_RESOURCES_DIR_NAME,
                                       CSAR_FILES_DIR_NAME)
        return CSARPackage._build_csar_resources(csar_files_path)

    @staticmethod
    def _build_csar_resources(basedir):
        result = []
        if os.path.isdir(basedir):
            for root, _, files in os.walk(os.path.abspath(basedir)):
                for f in files:
                    full_path = os.path.join(root, f)
                    relative_path = os.path.relpath(full_path, basedir)
                    result.append(relative_path)
        return result

    @staticmethod
    def _translate_constraint(constraint):
        if 'equal' in constraint:
            return CSARPackage._translate_equal_constraint(
                constraint['equal'])
        elif 'valid_values' in constraint:
            return CSARPackage._translate_valid_values_constraint(
                constraint['valid_values'])
        elif 'length' in constraint:
            return CSARPackage._translate_length_constraint(
                constraint['length'])
        elif 'in_range' in constraint:
            return CSARPackage._translate_range_constraint(
                constraint['in_range'])
        elif 'allowed_pattern' in constraint:
            return CSARPackage._translate_allowed_pattern_constraint(
                constraint['allowed_pattern'])

    @staticmethod
    def _translate_equal_constraint(value):
        return ".check($ == {0})".format(value)

    @staticmethod
    def _translate_allowed_pattern_constraint(value):
        return ".check(matches($, '{0}'))".format(value)

    @staticmethod
    def _translate_valid_values_constraint(values):
        return '.check($ in list({0}))'.format(
            ', '.join([CSARPackage._format_value(v) for v in values]))

    @staticmethod
    def _translate_length_constraint(value):
        if 'min' in value and 'max' in value:
            return '.check(len($) >= {0} and len($) <= {1})'.format(
                int(value['min']), int(value['max']))
        elif 'min' in value:
            return '.check(len($) >= {0})'.format(int(value['min']))
        elif 'max' in value:
            return '.check(len($) <= {0})'.format(int(value['max']))

    @staticmethod
    def _translate_range_constraint(value):
        if 'min' in value and 'max' in value:
            return '.check($ >= {0} and $ <= {1})'.format(
                int(value['min']), int(value['max']))
        elif 'min' in value:
            return '.check($ >= {0})'.format(int(value['min']))
        elif 'max' in value:
            return '.check($ <= {0})'.format(int(value['max']))

    @staticmethod
    def _format_value(value):
        if isinstance(value, str):
            return u"{}".format(value)
        return str(value)

    @staticmethod
    def _generate_workflow(csar, files):
        hot_files_map = {}
        for f in files:
            file_path = "$resources.string('{0}{1}')".format(
                CSAR_FILES_DIR_NAME, f)
            hot_files_map['../{0}'.format(f)] = YAQL(file_path)

        hot_env = YAQL("$.hotEnvironment")

        copy_outputs = []
        for key in (csar.get('outputs') or {}).keys():
            copy_outputs.append({YAQL('$.' + key): YAQL('$outputs.' + key)})

        deploy = [
            {YAQL('$environment'): YAQL(
                "$.find('io.murano.Environment').require()"
            )},
            {YAQL('$reporter'): YAQL(
                "new('io.murano.system.StatusReporter', "
                "environment => $environment)")},
            {
                'If': YAQL('$.getAttr(generatedHeatStackName) = null'),
                'Then': [
                    YAQL("$.setAttr(generatedHeatStackName, "
                         "'{0}_{1}'.format(randomName(), id($environment)))")
                ]
            },
            {YAQL('$stack'): YAQL(
                "new('io.murano.system.HeatStack', $environment, "
                "name => $.getAttr(generatedHeatStackName))")},

            YAQL("$reporter.report($this, "
                 "'Application deployment has started')"),

            {YAQL('$resources'): YAQL("new('io.murano.system.Resources')")},

            {YAQL('$template'): YAQL("$resources.yaml('template.yaml')")},
            YAQL('$stack.setTemplate($template)'),
            {YAQL('$parameters'): YAQL("$.templateParameters")},
            YAQL('$stack.setParameters($parameters)'),
            {YAQL('$files'): hot_files_map},
            YAQL('$stack.setFiles($files)'),
            {YAQL('$hotEnv'): hot_env},
            {
                'If': YAQL("bool($hotEnv)"),
                'Then': [
                    {YAQL('$envRelPath'): YAQL("'{0}' + $hotEnv".format(
                        CSAR_ENV_DIR_NAME))},
                    {YAQL('$hotEnvContent'): YAQL("$resources.string("
                                                  "$envRelPath)")},
                    YAQL('$stack.setHotEnvironment($hotEnvContent)')
                ]
            },

            YAQL("$reporter.report($this, 'Stack creation has started')"),
            {
                'Try': [YAQL('$stack.push()')],
                'Catch': [
                    {
                        'As': 'e',
                        'Do': [
                            YAQL("$reporter.report_error($this, $e.message)"),
                            {'Rethrow': None}
                        ]
                    }
                ],
                'Else': [
                    {YAQL('$outputs'): YAQL('$stack.output()')},
                    {'Do': copy_outputs},
                    YAQL("$reporter.report($this, "
                         "'Stack was successfully created')"),

                    YAQL("$reporter.report($this, "
                         "'Application deployment has finished')"),
                ]
            }
        ]

        destroy = [
            {YAQL('$environment'): YAQL(
                "$.find('io.murano.Environment').require()"
            )},
            {YAQL('$stack'): YAQL(
                "new('io.murano.system.HeatStack', $environment, "
                "name => $.getAttr(generatedHeatStackName))")},

            YAQL('$stack.delete()')
        ]

        return {
            'Workflow': {
                'deploy': {
                    'Body': deploy
                },
                'destroy': {
                    'Body': destroy
                }
            }
        }

    @staticmethod
    def _translate_ui_parameters(tosca, title):
        result_groups = []

        used_inputs = set()
        tosca_inputs = tosca.get('topology_template').get('inputs') or {}
        fields = []
        properties = []
        for input in tosca_inputs:
            input_value = tosca_inputs.get(input)
            if input_value:
                fields.append(CSARPackage._translate_ui_parameter(
                    input, input_value))
                used_inputs.add(input)
                properties.append(input)
        if fields or properties:
            result_groups.append((fields, properties))

        rest_group = []
        properties = []
        for key, value in tosca_inputs.items():
            if key not in used_inputs:
                rest_group.append(CSARPackage._translate_ui_parameter(
                    key, value))
                properties.append(key)
        if rest_group:
            result_groups.append((rest_group, properties))

        return result_groups

    @staticmethod
    def _translate_ui_parameter(name, parameter_spec):
        translated = {
            'name': name,
            'label': name.title().replace('_', ' ')
        }
        parameter_type = parameter_spec['type']
        if parameter_type == 'integer':
            translated['type'] = 'integer'
        elif parameter_type == 'boolean':
            translated['type'] = 'boolean'
        else:
            # string, json, and comma_delimited_list parameters are all
            # displayed as strings in UI. Any unsupported parameter would also
            # be displayed as strings.
            translated['type'] = 'string'

        label = parameter_spec.get('label')
        if label:
            translated['label'] = label

        if 'description' in parameter_spec:
            translated['description'] = parameter_spec['description']

        if 'default' in parameter_spec:
            translated['initial'] = parameter_spec['default']
            translated['required'] = False
        else:
            translated['required'] = True

        constraints = parameter_spec.get('constraints') or []
        translated_constraints = []

        for constraint in constraints:
            if 'length' in constraint:
                spec = constraint['length']
                if 'min' in spec:
                    translated['minLength'] = max(
                        translated.get('minLength', -sys.maxsize - 1),
                        int(spec['min']))
                if 'max' in spec:
                    translated['maxLength'] = min(
                        translated.get('maxLength', sys.maxsize),
                        int(spec['max']))

            elif 'range' in constraint:
                spec = constraint['range']
                if 'min' in spec and 'max' in spec:
                    ui_constraint = {
                        'expr': YAQL('$ >= {0} and $ <= {1}'.format(
                            spec['min'], spec['max']))
                    }
                elif 'min' in spec:
                    ui_constraint = {
                        'expr': YAQL('$ >= {0}'.format(spec['min']))
                    }
                else:
                    ui_constraint = {
                        'expr': YAQL('$ <= {0}'.format(spec['max']))
                    }
                if 'description' in constraint:
                    ui_constraint['message'] = constraint['description']
                translated_constraints.append(ui_constraint)

            elif 'valid_values' in constraint:
                values = constraint['valid_values']
                ui_constraint = {
                    'expr': YAQL('$ in list({0})'.format(', '.join(
                        [CSARPackage._format_value(v) for v in values])))
                }
                if 'description' in constraint:
                    ui_constraint['message'] = constraint['description']
                translated_constraints.append(ui_constraint)

            elif 'allowed_pattern' in constraint:
                pattern = constraint['allowed_pattern']
                ui_constraint = {
                    'expr': {
                        'regexpValidator': pattern
                    }
                }
                if 'description' in constraint:
                    ui_constraint['message'] = constraint['description']
                translated_constraints.append(ui_constraint)

        if translated_constraints:
            translated['validators'] = translated_constraints

        return translated

    @staticmethod
    def _generate_application_ui(groups, type_name, package_name=None,
                                 package_version=None):
        app = {
            '?': {
                'type': type_name
            }
        }
        if package_name:
            app['?']['package'] = package_name
        if package_version:
            app['?']['classVersion'] = package_version

        for i, record in enumerate(groups):
            section = app.setdefault('templateParameters', {})

            for property_name in record[1]:
                section[property_name] = YAQL(
                    '$.group{0}.{1}'.format(i, property_name))

        return app

    def _translate_ui(self):
        tosca = csar.CSAR(os.path.join(self._source_directory, 'csar.zip'))\
            .get_main_template_yaml()

        groups = CSARPackage._translate_ui_parameters(tosca, self.description)
        forms = []
        for i, record in enumerate(groups):
            forms.append({'group{0}'.format(i): {'fields': record[0]}})

        translated = {
            'Version': 2.2,
            'Application': CSARPackage._generate_application_ui(
                groups, self.full_name, self.full_name, str(self.version)),
            'Forms': forms
        }
        return yaml.dump(translated, Dumper=Dumper, default_style='"')



# Copyright (c) 2015 Mirantis, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from glance.common.glare import definitions


class MuranoPackage(definitions.ArtifactType):
    __endpoint__ = 'murano'

    type = definitions.String(allowed_values=['Application', 'Library'],
                              required=True,
                              mutable=False)

    author = definitions.String(required=False, mutable=False)
    display_name = definitions.String(required=True, mutable=True)
    enabled = definitions.Boolean(default=True)

    categories = definitions.Array(default=[], mutable=True)
    class_definitions = definitions.Array(unique=True, default=[],
                                          mutable=False)
    inherits = definitions.Dict(default={}, properties=definitions.Array(),
                                mutable=False)
    keywords = definitions.Array(default=[], mutable=True)
    logo = definitions.BinaryObject()
    archive = definitions.BinaryObject()
    ui_definition = definitions.BinaryObject()



#    Copyright (c) 2016 AT&T
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from murano.dsl import murano_object
from murano.engine.system import metadef_browser
from murano.tests.unit import base


class TestMetadefBrowser(base.MuranoTestCase):
    def setUp(self):
        super(TestMetadefBrowser, self).setUp()
        self.this = mock.MagicMock()

        self.glance_client_mock = mock.MagicMock()
        metadef_browser.MetadefBrowser._get_client = mock.MagicMock(
            return_value=self.glance_client_mock)
        metadef_browser.MetadefBrowser._client = mock.MagicMock(
            return_value=self.glance_client_mock)

        self.mock_object = mock.Mock(spec=murano_object.MuranoObject)

    @mock.patch("murano.dsl.helpers.get_execution_session")
    def test_get_objects(self, execution_session):
        namespace = None
        self.metadef = metadef_browser.MetadefBrowser(self.this)
        self.assertIsNotNone(self.metadef.get_objects(namespace))
        self.assertTrue(execution_session.called)
        self.assertTrue(metadef_browser.MetadefBrowser.
                        _client.metadefs_object.list.called)

    @mock.patch("murano.dsl.helpers.get_execution_session")
    def test_get_namespaces(self, execution_session):
        self.metadef = metadef_browser.MetadefBrowser(self.this)
        resource_type = self.mock_object
        self.assertIsNotNone(self.metadef.get_namespaces(resource_type))
        self.assertTrue(execution_session.called)
        self.assertTrue(metadef_browser.MetadefBrowser.
                        _client.metadefs_namespace.list.called)



# Copyright (c) 2016 AT&T
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest import mock

from oslo_config import cfg

from murano.dsl import murano_method
from murano.dsl import murano_type
from murano.engine.system import net_explorer
from murano.tests.unit import base

CONF = cfg.CONF


class TestNetExplorer(base.MuranoTestCase):
    def setUp(self):
        super(TestNetExplorer, self).setUp()
        self.mock_class = mock.MagicMock(spec=murano_type.MuranoClass)
        self.mock_method = mock.MagicMock(spec=murano_method.MuranoMethod)

        self._this = mock.MagicMock()
        self.region_name = "test-region"

        self.addCleanup(mock.patch.stopall)

    @mock.patch("murano.engine.system.net_explorer.nclient")
    @mock.patch("murano.engine.system.net_explorer.auth_utils")
    @mock.patch("murano.dsl.helpers.get_execution_session")
    def test_get_available_cidr(self, execution_session,
                                mock_authentication, mock_nclient):
        ne = net_explorer.NetworkExplorer(self._this, self.region_name)
        router_id = 12
        net_id = 144
        self.assertIsNotNone(ne.get_available_cidr(router_id, net_id))
        self.assertTrue(execution_session.called)

    @mock.patch("murano.engine.system.net_explorer.nclient")
    @mock.patch("murano.engine.system.net_explorer.auth_utils")
    @mock.patch("murano.dsl.helpers.get_execution_session")
    def test_list(self, execution_session, mock_authentication, mock_nclient):
        ne = net_explorer.NetworkExplorer(self._this, self.region_name)
        self.assertEqual(ne.list_networks(),
                         ne._client.list_networks()['networks'])
        self.assertEqual(ne.list_subnetworks(),
                         ne._client.list_subnets()['subnets'])
        self.assertEqual(ne.list_ports(), ne._client.list_ports()['ports'])
        self.assertEqual(ne.list_neutron_extensions(),
                         ne._client.list_extensions()['extensions'])
        self.assertEqual(ne.get_default_dns(), ne._settings.default_dns)

    @mock.patch("murano.engine.system.net_explorer.nclient")
    @mock.patch("murano.engine.system.net_explorer.auth_utils")
    @mock.patch("murano.dsl.helpers.get_execution_session")
    def test_get_router_error(self, execution_session,
                              mock_authentication, mock_nclient):
        ne = net_explorer.NetworkExplorer(self._this, self.region_name)
        self.assertRaises(KeyError, ne.get_default_router)

    @mock.patch("murano.engine.system.net_explorer.nclient")
    @mock.patch("murano.engine.system.net_explorer.auth_utils")
    @mock.patch("murano.dsl.helpers.get_execution_session")
    def test_get_ext_network_id_router(self, execution_session,
                                       mock_authentication, mock_nclient):
        ne = net_explorer.NetworkExplorer(self._this, self.region_name)
        router_id = 12
        self.assertIsNone(ne.get_external_network_id_for_router(router_id))

    @mock.patch("murano.engine.system.net_explorer.nclient")
    @mock.patch("murano.engine.system.net_explorer.auth_utils")
    @mock.patch("murano.dsl.helpers.get_execution_session")
    def test_get_ext_network_id_network(self, execution_session,
                                        mock_authentication, mock_nclient):
        ne = net_explorer.NetworkExplorer(self._this, self.region_name)
        net_id = 144
        self.assertEqual(net_id,
                         ne.get_external_network_id_for_network(net_id))

    @mock.patch("murano.engine.system.net_explorer.nclient")
    @mock.patch("murano.engine.system.net_explorer.auth_utils")
    @mock.patch("murano.dsl.helpers.get_execution_session")
    def test_get_cidr_none_router(self, execution_session,
                                  mock_authentication, mock_nclient):
        ne = net_explorer.NetworkExplorer(self._this, self.region_name)
        router_id = None
        self.assertEqual([], ne._get_cidrs_taken_by_router(router_id))



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from murano.common.helpers import token_sanitizer
from murano.db import models
from murano.services import states


def get_environment(session, unit):
    environment = unit.query(models.Environment).get(
        session.environment_id)
    return environment


def update_task(action, session, task, unit):
    objects = session.description.get('Objects', None)
    session.state = states.SessionState.DELETING if objects is None \
        else states.SessionState.DEPLOYING
    task_info = models.Task()
    task_info.environment_id = session.environment_id
    if objects:
        task_info.description = token_sanitizer.TokenSanitizer().sanitize(
            dict(session.description.get('Objects')))
    task_info.action = task['action']
    status = models.Status()
    status.text = 'Action {0} is scheduled'.format(action)
    status.level = 'info'
    task_info.statuses.append(status)
    with unit.begin():
        unit.add(session)
        unit.add(task_info)
    return task_info.id



#    Copyright (c) 2016 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc


class ContractMethod(object):
    @abc.abstractmethod
    def transform(self):
        raise NotImplementedError()

    @abc.abstractmethod
    def validate(self):
        raise NotImplementedError()

    def finalize(self):
        return self.value

    def check_type(self):
        return self.validate()

    def generate_schema(self):
        return self.value

    def __getattr__(self, item):
        return self.context[item]


class ObjRef(object):
    def __init__(self, object_id):
        self.object_id = object_id



# Copyright (c) 2014 Hewlett-Packard Development Company, L.P.
# Copyright (c) 2016 AT&T Corp
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from unittest import mock

from oslo_utils import timeutils
from webob import exc

from murano.api.v1 import actions
from murano.common import policy
from murano.db import models
from murano.db import session as db_session
from murano.services import states
import murano.tests.unit.api.base as tb
import murano.tests.unit.utils as test_utils


@mock.patch.object(policy, 'check')
class TestActionsApi(tb.ControllerTest, tb.MuranoApiTestCase):

    def setUp(self):
        super(TestActionsApi, self).setUp()
        self.controller = actions.Controller()

    def test_execute_action(self, mock_policy_check):
        """Test that action execution results in the correct rpc call."""
        self._set_policy_rules(
            {'execute_action': '@'}
        )

        fake_now = timeutils.utcnow()
        expected = dict(
            id='12345',
            name='my-env',
            version=0,
            created=fake_now,
            updated=fake_now,
            tenant_id=self.tenant,
            description={
                'Objects': {
                    '?': {'id': '12345',
                          '_actions': {
                              'actionsID_action': {
                                  'enabled': True,
                                  'name': 'Testaction'
                              }
                          }}
                },
                'Attributes': {}
            }
        )
        e = models.Environment(**expected)
        test_utils.save_models(e)

        rpc_task = {
            'action': {
                'args': '{}',
                'method': 'Testaction',
                'object_id': '12345'
            },
            'project_id': self.tenant,
            'user_id': self.user,
            'model': {
                'Attributes': {},
                'Objects': {
                    'applications': [],
                    '?': {
                        '_actions': {
                            'actionsID_action': {
                                'enabled': True,
                                'name': 'Testaction'
                            }
                        },
                        'id': '12345'
                    }
                }
            },
            'token': None,
            'id': '12345'
        }

        req = self._post('/environments/12345/actions/actionID_action', b'{}')
        result = self.controller.execute(req, '12345', 'actionsID_action',
                                         '{}')

        self.mock_engine_rpc.handle_task.assert_called_once_with(rpc_task)

        self.assertIn('task_id', result)

    def _create_session_with_state(self, environment, user_id, state):
        unit = db_session.get_session()
        session = models.Session()
        session.environment_id = environment.id
        session.user_id = user_id
        session.state = state
        session.version = environment.version
        with unit.begin():
            unit.add(session)

    def test_execute_action_with_session_in_deploying_state(self, _):
        """Test whether session in the deploying state throws error."""
        self._set_policy_rules(
            {'execute_action': '@'}
        )

        fake_now = timeutils.utcnow()
        expected = dict(
            id='12345',
            name='my-env',
            version=0,
            created=fake_now,
            updated=fake_now,
            tenant_id=self.tenant,
            description={
                'Objects': {
                    '?': {'id': '12345',
                          '_actions': {
                              'actionsID_action': {
                                  'enabled': True,
                                  'name': 'Testaction'
                              }
                          }}
                },
                'Attributes': {}
            }
        )
        environment = models.Environment(**expected)
        test_utils.save_models(environment)
        req = self._post('/environments/12345/actions/actionID_action', b'{}')
        user_id = req.context.user

        self._create_session_with_state(environment, user_id,
                                        states.SessionState.DEPLOYING)
        self.assertRaises(exc.HTTPForbidden, self.controller.execute,
                          req, '12345', 'actionsID_action', {})

    def test_execute_action_with_session_in_deleting_state(self, _):
        """Test whether session in deleting state throws error."""
        self._set_policy_rules(
            {'execute_action': '@'}
        )

        fake_now = timeutils.utcnow()
        expected = dict(
            id='12345',
            name='my-env',
            version=0,
            created=fake_now,
            updated=fake_now,
            tenant_id=self.tenant,
            description={
                'Objects': {
                    '?': {'id': '12345',
                          '_actions': {
                              'actionsID_action': {
                                  'enabled': True,
                                  'name': 'Testaction'
                              }
                          }}
                },
                'Attributes': {}
            }
        )
        environment = models.Environment(**expected)
        test_utils.save_models(environment)
        req = self._post('/environments/12345/actions/actionID_action', b'{}')
        user_id = req.context.user

        self._create_session_with_state(environment, user_id,
                                        states.SessionState.DELETING)
        self.assertRaises(exc.HTTPForbidden, self.controller.execute,
                          req, '12345', 'actionsID_action', {})

    @mock.patch('murano.db.services.sessions.SessionServices.validate')
    def test_execute_action_with_invalid_session_version(self, mocked_function,
                                                         _):
        """Test whether validate session function throws error."""
        self._set_policy_rules(
            {'execute_action': '@'}
        )

        fake_now = timeutils.utcnow()
        expected = dict(
            id='12345',
            name='my-env',
            version=0,
            created=fake_now,
            updated=fake_now,
            tenant_id=self.tenant,
            description={
                'Objects': {
                    '?': {'id': '12345',
                          '_actions': {
                              'actionsID_action': {
                                  'enabled': True,
                                  'name': 'Testaction'
                              }
                          }}
                },
                'Attributes': {}
            }
        )
        environment = models.Environment(**expected)
        test_utils.save_models(environment)
        req = self._post('/environments/12345/actions/actionID_action', b'{}')

        mocked_function.return_value = False
        self.assertRaises(exc.HTTPForbidden, self.controller.execute,
                          req, '12345', 'actionsID_action', {})

    def test_get_result(self, _):
        """Result of task with given id and environment id is returned."""
        now = timeutils.utcnow()
        expected_environment_id = 'test_environment'
        expected_task_id = 'test_task'
        expected_result = {'test_result': 'test_result'}

        environment = models.Environment(
            id=expected_environment_id,
            name='test_environment', created=now, updated=now,
            tenant_id=self.tenant
        )

        task = models.Task(
            id=expected_task_id,
            started=now,
            finished=now,
            result=expected_result,
            environment_id=expected_environment_id
        )

        test_utils.save_models(environment, task)

        request = self._get(
            '/environments/{environment_id}/actions/{task_id}'
            .format(environment_id=expected_environment_id,
                    task_id=expected_task_id),
        )

        response = request.get_response(self.api)

        self.assertEqual(200, response.status_code)
        self.assertEqual(expected_result, response.json)

    def test_get_result_not_found(self, _):
        """Return 404 on null task

        If task does not exist, it should be handled correctly
        and API should return 404.
        """
        expected_environment_id = 'test_environment'

        environment = models.Environment(
            id=expected_environment_id,
            name='test_environment',
            tenant_id=self.tenant
        )
        test_utils.save_models(environment)

        request = self._get(
            '/environments/{environment_id}/actions/{task_id}'
            .format(environment_id=expected_environment_id,
                    task_id='not_existent_task_id'),
        )

        response = request.get_response(self.api)

        self.assertEqual(404, response.status_code)



# Copyright (c) 2014 Hewlett-Packard Development Company, L.P.
# Copyright (c) 2016 AT&T Corp
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from heatclient.v1 import stacks
from unittest import mock

from oslo_config import cfg

from murano.engine.system import heat_stack
from murano.tests.unit import base

CLS_NAME = 'murano.engine.system.heat_stack.HeatStack'
CONF = cfg.CONF


class TestHeatStack(base.MuranoTestCase):
    def setUp(self):
        super(TestHeatStack, self).setUp()
        self.heat_client_mock = mock.Mock()
        self.heat_client_mock.stacks = mock.MagicMock(spec=stacks.StackManager)
        self.override_config('stack_tags', ['test-murano'], 'heat')
        self.mock_tag = ','.join(CONF.heat.stack_tags)
        self._patch_get_client()

    def tearDown(self):
        super(TestHeatStack, self).tearDown()
        self.addCleanup(mock.patch.stopall)

    def _patch_get_client(self):
        self.get_client_patcher = mock.patch(
            'murano.engine.system.heat_stack.HeatStack._get_client',
            return_value=self.heat_client_mock)
        self.get_token_client_patcher = mock.patch.object(
            heat_stack.HeatStack, '_get_token_client',
            return_value=self.heat_client_mock)
        self.get_client_patcher.start()
        self.get_token_client_patcher.start()

    def _unpatch_get_client(self):
        self.get_client_patcher.stop()
        self.get_token_client_patcher.stop()

    @mock.patch(CLS_NAME + '._wait_state')
    @mock.patch(CLS_NAME + '._get_status')
    def test_push_adds_version(self, status_get, wait_st):
        """Assert that if heat_template_version is omitted, it's added."""

        status_get.return_value = 'NOT_FOUND'
        wait_st.return_value = {}
        hs = heat_stack.HeatStack('test-stack', 'Generated by TestHeatStack')
        hs._template = {'resources': {'test': 1}}
        hs._files = {}
        hs._hot_environment = {}
        hs._parameters = {}
        hs._applied = False
        hs.push()

        hs = heat_stack.HeatStack(
            'test-stack', 'Generated by TestHeatStack')
        hs._template = {'resources': {'test': 1}}
        hs._files = {}
        hs._parameters = {}
        hs._applied = False
        hs.push()

        expected_template = {
            'heat_template_version': '2013-05-23',
            'description': 'Generated by TestHeatStack',
            'resources': {'test': 1}
        }
        self.heat_client_mock.stacks.create.assert_called_with(
            stack_name='test-stack',
            disable_rollback=True,
            parameters={},
            template=expected_template,
            files={},
            environment={},
            tags=self.mock_tag
        )
        self.assertTrue(hs._applied)

    @mock.patch(CLS_NAME + '._wait_state')
    @mock.patch(CLS_NAME + '._get_status')
    def test_description_is_optional(self, status_get, wait_st):
        """Assert that if heat_template_version is omitted, it's added."""

        status_get.return_value = 'NOT_FOUND'
        wait_st.return_value = {}
        hs = heat_stack.HeatStack('test-stack', None)
        hs._template = {'resources': {'test': 1}}
        hs._files = {}
        hs._hot_environment = ''
        hs._parameters = {}
        hs._applied = False
        hs.push()

        expected_template = {
            'heat_template_version': '2013-05-23',
            'resources': {'test': 1}
        }
        self.heat_client_mock.stacks.create.assert_called_with(
            stack_name='test-stack',
            disable_rollback=True,
            parameters={},
            template=expected_template,
            files={},
            environment='',
            tags=self.mock_tag
        )
        self.assertTrue(hs._applied)

    @mock.patch(CLS_NAME + '._wait_state')
    @mock.patch(CLS_NAME + '._get_status')
    def test_heat_files_are_sent(self, status_get, wait_st):
        """Assert that if heat_template_version is omitted, it's added."""

        status_get.return_value = 'NOT_FOUND'
        wait_st.return_value = {}
        hs = heat_stack.HeatStack('test-stack', None)
        hs._description = None
        hs._template = {'resources': {'test': 1}}
        hs._files = {"heatFile": "file"}
        hs._hot_environment = ''
        hs._parameters = {}
        hs._applied = False
        hs.push()

        expected_template = {
            'heat_template_version': '2013-05-23',
            'resources': {'test': 1}
        }
        self.heat_client_mock.stacks.create.assert_called_with(
            stack_name='test-stack',
            disable_rollback=True,
            parameters={},
            template=expected_template,
            files={"heatFile": "file"},
            environment='',
            tags=self.mock_tag
        )
        self.assertTrue(hs._applied)

    @mock.patch(CLS_NAME + '._wait_state')
    @mock.patch(CLS_NAME + '._get_status')
    def test_heat_environments_are_sent(self, status_get, wait_st):
        """Assert that if heat_template_version is omitted, it's added."""

        status_get.return_value = 'NOT_FOUND'
        wait_st.return_value = {}
        hs = heat_stack.HeatStack('test-stack', None)
        hs._description = None
        hs._template = {'resources': {'test': 1}}
        hs._files = {"heatFile": "file"}
        hs._hot_environment = 'environments'
        hs._parameters = {}
        hs._applied = False
        hs.push()

        expected_template = {
            'heat_template_version': '2013-05-23',
            'resources': {'test': 1}
        }
        self.heat_client_mock.stacks.create.assert_called_with(
            stack_name='test-stack',
            disable_rollback=True,
            parameters={},
            template=expected_template,
            files={"heatFile": "file"},
            environment='environments',
            tags=self.mock_tag
        )
        self.assertTrue(hs._applied)

    @mock.patch(CLS_NAME + '._wait_state')
    @mock.patch(CLS_NAME + '._get_status')
    def test_heat_async_push(self, status_get, wait_st):
        """Assert that if heat_template_version is omitted, it's added."""

        status_get.return_value = 'NOT_FOUND'
        wait_st.return_value = {}
        hs = heat_stack.HeatStack('test-stack', None)
        hs._description = None
        hs._template = {'resources': {'test': 1}}
        hs._files = {"heatFile": "file"}
        hs._hot_environment = 'environments'
        hs._parameters = {}
        hs._applied = False
        with mock.patch('murano.dsl.dsl.get_execution_session'):
            hs.push(is_async=True)

        expected_template = {
            'heat_template_version': '2013-05-23',
            'resources': {'test': 1}
        }
        self.heat_client_mock.stacks.create.assert_not_called()
        hs.output()
        self.heat_client_mock.stacks.create.assert_called_with(
            stack_name='test-stack',
            disable_rollback=True,
            parameters={},
            template=expected_template,
            files={"heatFile": "file"},
            environment='environments',
            tags=self.mock_tag
        )
        self.assertTrue(hs._applied)

    @mock.patch(CLS_NAME + '._wait_state')
    @mock.patch(CLS_NAME + '._get_status')
    @mock.patch.object(heat_stack, 'LOG')
    def test_push_except_http_conflict(self, mock_log, status_get, wait_st):
        status_get.return_value = 'NOT_FOUND'
        wait_st.return_value = {}
        hs = heat_stack.HeatStack('test-stack', None)
        hs._description = None
        hs._template = {'resources': {'test': 1}}
        hs._files = {"heatFile": "file"}
        hs._hot_environment = 'environments'
        hs._parameters = {}
        hs._applied = False
        hs._get_token_client().stacks.create.side_effect = [
            heat_stack.heat_exc.HTTPConflict('test_error_msg'),
            None
        ]

        hs.push()

        mock_log.warning.assert_called_with(
            'Conflicting operation: ERROR: test_error_msg')

    @mock.patch(CLS_NAME + '.current')
    def test_update_wrong_template_version(self, current):
        """Template version other than expected should cause error."""

        hs = heat_stack.HeatStack(
            'test-stack', 'Generated by TestHeatStack')
        hs._template = {'resources': {'test': 1}}

        invalid_template = {
            'heat_template_version': 'something else'
        }

        current.return_value = {}

        e = self.assertRaises(heat_stack.HeatStackError,
                              hs.update_template,
                              invalid_template)
        err_msg = "Currently only heat_template_version 2013-05-23 "\
                  "is supported."
        self.assertEqual(err_msg, str(e))

        # Check it's ok without a version
        hs.update_template({})
        expected = {'resources': {'test': 1}}
        self.assertEqual(expected, hs._template)

        # .. or with a good version
        hs.update_template({'heat_template_version': '2013-05-23'})
        expected['heat_template_version'] = '2013-05-23'
        self.assertEqual(expected, hs._template)

    @mock.patch(CLS_NAME + '._wait_state')
    @mock.patch(CLS_NAME + '._get_status')
    def test_heat_stack_tags_are_sent(self, status_get, wait_st):
        """Assert heat_stack tags are sent

        Assert that heat_stack `tags` parameter get push & with
        value from config parameter `stack_tags`.
        """

        status_get.return_value = 'NOT_FOUND'
        wait_st.return_value = {}
        self.override_config('stack_tags', ['test-murano', 'murano-tag'],
                             'heat')
        hs = heat_stack.HeatStack('test-stack', None)
        hs._description = None
        hs._template = {'resources': {'test': 1}}
        hs._files = {}
        hs._hot_environment = ''
        hs._parameters = {}
        hs._applied = False
        hs._tags = ','.join(CONF.heat.stack_tags)
        hs.push()

        expected_template = {
            'heat_template_version': '2013-05-23',
            'resources': {'test': 1}
        }
        self.heat_client_mock.stacks.create.assert_called_with(
            stack_name='test-stack',
            disable_rollback=True,
            parameters={},
            template=expected_template,
            files={},
            environment='',
            tags=','.join(CONF.heat.stack_tags)
        )
        self.assertTrue(hs._applied)

    @mock.patch(CLS_NAME + '._wait_state')
    @mock.patch(CLS_NAME + '._get_status')
    def test_parameters(self, status_get, wait_st):
        status_get.return_value = 'NOT_FOUND'
        wait_st.return_value = {}
        self.override_config('stack_tags', ['test-murano', 'murano-tag'],
                             'heat')
        hs = heat_stack.HeatStack('test-stack', None)
        hs._description = None
        hs._template = {'resources': {'test': 1}}
        hs._files = {}
        hs._hot_environment = ''
        hs._parameters = {}
        hs._applied = False
        hs._tags = ','.join(CONF.heat.stack_tags)
        hs.push()

        expected_template = {
            'heat_template_version': '2013-05-23',
            'resources': {'test': 1}
        }
        self.heat_client_mock.stacks.create.assert_called_with(
            stack_name='test-stack',
            disable_rollback=True,
            parameters={},
            template=expected_template,
            files={},
            environment='',
            tags=','.join(CONF.heat.stack_tags)
        )

        self.assertEqual(hs.parameters(), hs._parameters)

    @mock.patch(CLS_NAME + '._wait_state')
    @mock.patch(CLS_NAME + '._get_status')
    def test_reload(self, status_get, wait_st):
        status_get.return_value = 'NOT_FOUND'
        wait_st.return_value = {}
        self.override_config('stack_tags', ['test-murano', 'murano-tag'],
                             'heat')
        hs = heat_stack.HeatStack('test-stack', None)
        hs._description = None
        hs._template = {'resources': {'test': 1}}
        hs._files = {}
        hs._hot_environment = ''
        hs._parameters = {}
        hs._applied = False
        hs._tags = ','.join(CONF.heat.stack_tags)
        hs.push()

        expected_template = {
            'heat_template_version': '2013-05-23',
            'resources': {'test': 1}
        }
        self.heat_client_mock.stacks.create.assert_called_with(
            stack_name='test-stack',
            disable_rollback=True,
            parameters={},
            template=expected_template,
            files={},
            environment='',
            tags=','.join(CONF.heat.stack_tags)
        )

        hs.reload()
        stack_info = self.heat_client_mock.stacks.get(stack_id=hs._name)
        self.assertEqual(hs._template, hs._client.stacks.template(
            stack_id='{0}/{1}'.format(
                stack_info.stack_name,
                stack_info.id)))

    @mock.patch(CLS_NAME + '._wait_state')
    @mock.patch(CLS_NAME + '._get_status')
    def test_delete(self, status_get, wait_st):
        status_get.return_value = 'NOT_FOUND'
        wait_st.return_value = {}
        self.override_config('stack_tags', ['test-murano', 'murano-tag'],
                             'heat')
        hs = heat_stack.HeatStack('test-stack', None)
        hs._description = None
        hs._template = {'resources': {'test': 1}}
        hs._files = {}
        hs._hot_environment = ''
        hs._parameters = {}
        hs._applied = False
        hs._tags = ','.join(CONF.heat.stack_tags)
        hs.push()

        expected_template = {
            'heat_template_version': '2013-05-23',
            'resources': {'test': 1}
        }
        self.heat_client_mock.stacks.create.assert_called_with(
            stack_name='test-stack',
            disable_rollback=True,
            parameters={},
            template=expected_template,
            files={},
            environment='',
            tags=','.join(CONF.heat.stack_tags)
        )

        hs.delete()
        self.assertEqual({}, hs._template)
        self.assertTrue(hs._applied)

    @mock.patch(CLS_NAME + '._wait_state')
    @mock.patch(CLS_NAME + '._get_status')
    @mock.patch.object(heat_stack, 'LOG')
    def test_delete_except_not_found(self, mock_log, status_get, wait_st):
        status_get.return_value = 'NOT_FOUND'
        wait_st.return_value = {}

        hs = heat_stack.HeatStack('test-stack', None)
        hs._template = {'resources': {'test': 1}}
        hs._files = {}
        hs._hot_environment = ''
        hs._parameters = {}
        hs._applied = False
        hs.push()

        hs._client.stacks.delete.side_effect =\
            heat_stack.heat_exc.NotFound
        hs.delete()

        self.assertTrue(hs._applied)
        self.assertEqual({}, hs._template)
        mock_log.warning.assert_called_with(
            'Stack test-stack already deleted?')

    @mock.patch(CLS_NAME + '._wait_state')
    @mock.patch(CLS_NAME + '._get_status')
    @mock.patch.object(heat_stack, 'LOG')
    def test_delete_except_http_conflict(self, mock_log, status_get, wait_st):
        status_get.return_value = 'NOT_FOUND'
        wait_st.return_value = {}

        hs = heat_stack.HeatStack('test-stack', None)
        hs._template = {'resources': {'test': 1}}
        hs._files = {}
        hs._hot_environment = ''
        hs._parameters = {}
        hs._applied = False
        hs.push()

        hs._client.stacks.delete.side_effect = [
            heat_stack.heat_exc.HTTPConflict('test_error_msg'),
            None
        ]
        hs.delete()

        self.assertTrue(hs._applied)
        self.assertEqual({}, hs._template)
        mock_log.warning.assert_called_with('Conflicting operation: '
                                            'ERROR: test_error_msg')

    @mock.patch(CLS_NAME + '._wait_state')
    @mock.patch(CLS_NAME + '._get_status')
    def test_set_template_and_params(self, status_get, wait_st):
        status_get.return_value = 'NOT_FOUND'
        wait_st.return_value = {}
        self.override_config('stack_tags', ['test-murano', 'murano-tag'],
                             'heat')
        hs = heat_stack.HeatStack('test-stack', None)
        hs._description = None
        hs._template = {'resources': {'test': 1}}
        hs._files = {}
        hs._hot_environment = ''
        hs._parameters = {}
        hs._applied = False
        hs._tags = ','.join(CONF.heat.stack_tags)
        hs.push()

        expected_template = {
            'heat_template_version': '2013-05-23',
            'resources': {'test': 1}
        }
        self.heat_client_mock.stacks.create.assert_called_with(
            stack_name='test-stack',
            disable_rollback=True,
            parameters={},
            template=expected_template,
            files={},
            environment='',
            tags=','.join(CONF.heat.stack_tags)
        )

        new_template = {'resources': {'test': 2}}
        new_parameters = {'parameters': {'test': 1}}
        hs.set_template(new_template)
        self.assertEqual(new_template, hs._template)
        hs.set_parameters(new_parameters)
        self.assertEqual(new_parameters, hs._parameters)

    @mock.patch(CLS_NAME + '._wait_state')
    @mock.patch(CLS_NAME + '._get_status')
    def test_set_hot_env_and_files(self, status_get, wait_st):
        status_get.return_value = 'NOT_FOUND'
        wait_st.return_value = {}
        self.override_config('stack_tags', ['test-murano', 'murano-tag'],
                             'heat')
        hs = heat_stack.HeatStack('test-stack', None)
        hs._description = None
        hs._template = {'resources': {'test': 1}}
        hs._files = {}
        hs._hot_environment = ''
        hs._parameters = {}
        hs._applied = False
        hs._tags = ','.join(CONF.heat.stack_tags)
        hs.push()

        expected_template = {
            'heat_template_version': '2013-05-23',
            'resources': {'test': 1}
        }
        self.heat_client_mock.stacks.create.assert_called_with(
            stack_name='test-stack',
            disable_rollback=True,
            parameters={},
            template=expected_template,
            files={},
            environment='',
            tags=','.join(CONF.heat.stack_tags)
        )

        new_hot_env = 'test'
        new_files = {'files': {'test': 1}}
        hs.set_hot_environment(new_hot_env)
        self.assertEqual(new_hot_env, hs._hot_environment)
        hs.set_files(new_files)
        self.assertEqual(new_files, hs._files)

    @mock.patch(CLS_NAME + '._wait_state')
    @mock.patch(CLS_NAME + '._get_status')
    def test_none_template(self, status_get, wait_st):
        status_get.return_value = 'NOT_FOUND'
        wait_st.return_value = {}
        self.override_config('stack_tags', ['test-murano', 'murano-tag'],
                             'heat')
        hs = heat_stack.HeatStack('test-stack', None)
        hs._description = None
        hs._template = None
        hs._files = {}
        hs._hot_environment = ''
        hs._parameters = {}
        hs._applied = True
        hs._tags = ','.join(CONF.heat.stack_tags)
        self.assertIsNone(hs.push())

    @mock.patch(CLS_NAME + '._wait_state')
    def test_get_hot_status(self, wait_st):
        wait_st.return_value = {}
        self.override_config('stack_tags', ['test-murano', 'murano-tag'],
                             'heat')
        hs = heat_stack.HeatStack('test-stack', None)
        hs._description = None
        hs._template = {'resources': {'test': 1}}
        hs._files = {}
        hs._hot_environment = ''
        hs._parameters = {}
        hs._applied = False
        hs._tags = ','.join(CONF.heat.stack_tags)
        hs.push()

        self.assertIsNone(hs._get_status())
        self.assertTrue(wait_st.called)
        self.assertEqual({}, hs.output())

    def test_current_except_http_notfound(self):
        hs = heat_stack.HeatStack(
            'test-stack', 'Generated by TestHeatStack')
        hs._template = None
        hs._applied = False
        hs._parameters = {'param1': 'val1', 'param2': 'val2'}
        hs._client.stacks.get.side_effect = heat_stack.heat_exc.HTTPNotFound

        current = hs.current()

        self.assertEqual({}, current)
        self.assertEqual(True, hs._applied)
        self.assertEqual({}, hs._template)
        self.assertEqual({}, hs._parameters)

    @mock.patch.object(heat_stack, 'auth_utils')
    def test_get_client(self, mock_auth_utils):
        self._unpatch_get_client()

        mock_auth_utils.get_session_client_parameters.return_value =\
            {'endpoint': 'test_endpoint/v1'}

        client = heat_stack.HeatStack._get_client('test_region_name')

        self.assertIsNotNone(client)
        self.assertEqual("<class 'heatclient.v1.client.Client'>",
                         str(client.__class__))
        mock_auth_utils.get_client_session.assert_called_with(
            conf='heat')

    @mock.patch.object(heat_stack, 'auth_utils')
    def test_get_token_client(self, mock_auth_utils):
        self._unpatch_get_client()

        mock_auth_utils.get_session_client_parameters.return_value =\
            {'endpoint': 'test_endpoint/v1'}

        hs = heat_stack.HeatStack('test-stack', 'Generated by TestHeatStack')

        token_client = hs._get_token_client()

        self.assertIsNotNone(token_client)
        self.assertEqual("<class 'heatclient.v1.client.Client'>",
                         str(token_client.__class__))
        mock_auth_utils.get_token_client_session.assert_called_with(
            conf='heat')

    def test_wait_state(self):
        hs = heat_stack.HeatStack('test-stack', None)
        hs._client.stacks.get.return_value =\
            mock.Mock(stack_status='CREATE_COMPLETE')

        result = hs._wait_state(lambda status: status == 'CREATE_COMPLETE')

        self.assertEqual({}, result)

    def test_wait_state_with_outputs(self):
        hs = heat_stack.HeatStack('test-stack', None)
        hs._client.stacks.get.side_effect = [
            mock.Mock(stack_status='IN_PROGRESS'),
            mock.Mock(stack_status='CREATE_COMPLETE',
                      outputs=[{'output_key': 'key1', 'output_value': 'val1'},
                               {'output_key': 'key2', 'output_value': 'val2'}])
        ]

        result = hs._wait_state(lambda status: status == 'CREATE_COMPLETE')

        self.assertNotEqual({}, result)
        self.assertEqual({'key1': 'val1', 'key2': 'val2'}, result)

    def test_wait_state_with_multiple_states(self):
        """Test that only the first state is checked."""

        hs = heat_stack.HeatStack('test-stack', None)
        hs._client.stacks.get.side_effect = [
            mock.Mock(stack_status=['IN_PROGRESS', 'NOT_FOUND']),
            mock.Mock(stack_status='CREATE_COMPLETE')
        ]

        result = hs._wait_state(lambda status: status == 'CREATE_COMPLETE')

        self.assertEqual({}, result)

    @mock.patch.object(heat_stack, 'eventlet')
    def test_wait_state_with_wait_progress_true(self, mock_eventlet):
        hs = heat_stack.HeatStack('test-stack', None)
        hs._last_stack_timestamps = ('creation_time', 'updated_time')
        hs._client.stacks.get.side_effect = [
            mock.Mock(stack_status='TEST_STATUS',
                      creation_time='creation_time',
                      updated_time='updated_time'),
            mock.Mock(stack_status='TEST_STATUS',
                      creation_time='creation_time',
                      updated_time='updated_time'),
            mock.Mock(stack_status='CREATE_COMPLETE')
        ]

        result = hs._wait_state(lambda status: status == 'CREATE_COMPLETE',
                                wait_progress=True)

        self.assertEqual({}, result)
        self.assertEqual(3, hs._client.stacks.get.call_count)
        self.assertEqual(2, mock_eventlet.sleep.call_count)
        expected_calls = [mock.call.sleep(2), mock.call.sleep(2)]
        self.assertEqual(expected_calls, mock_eventlet.sleep.mock_calls)

    def test_wait_state_except_http_not_found(self):
        hs = heat_stack.HeatStack('test-stack', None)
        hs._client.stacks.get.side_effect = heat_stack.heat_exc.HTTPNotFound

        # If NOT FOUND is the expected status, then should run successfully.
        result = hs._wait_state(lambda status: status == 'NOT_FOUND')
        self.assertEqual({}, result)

        # Else EnvironmentError should be thrown.
        expected_error_msg = "Unexpected stack state {0}"\
            .format('NOT_FOUND')
        with self.assertRaisesRegex(EnvironmentError,
                                    expected_error_msg):
            hs._wait_state(lambda status: status == 'CREATE_COMPLETE')

    @mock.patch.object(heat_stack, 'eventlet')
    def test_wait_state_except_general_exception(self, mock_eventlet):
        """Test whether 4 tries are executed before exception raised."""

        hs = heat_stack.HeatStack('test-stack', None)
        hs._client.stacks.get.side_effect = Exception('test_exception_msg')

        with self.assertRaisesRegex(Exception,
                                    'test_exception_msg'):
            hs._wait_state(lambda status: status == 'CREATE_COMPLETE')

        expected_calls = [mock.call.sleep(2), mock.call.sleep(4),
                          mock.call.sleep(8)]
        self.assertEqual(4, hs._client.stacks.get.call_count)
        self.assertEqual(3, mock_eventlet.sleep.call_count)
        self.assertEqual(expected_calls, mock_eventlet.sleep.mock_calls)

    def test_wait_state_except_environment_error(self):
        hs = heat_stack.HeatStack('test-stack', None)
        hs._client.stacks.get.side_effect = [
            mock.Mock(stack_status='IN_PROGRESS'),
            mock.Mock(stack_status='UNEXPECTED_STATUS',
                      stack_status_reason='test_reason')
        ]

        expected_error_msg = "Unexpected stack state {0}: {1}"\
            .format('UNEXPECTED_STATUS', 'test_reason')
        with self.assertRaisesRegex(EnvironmentError,
                                    expected_error_msg):
            hs._wait_state(lambda status: status == 'CREATE_COMPLETE')



#    Copyright (c) 2015, Telefonica I+D.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_db import exception as db_exc
from oslo_log import log as logging
from webob import exc

from murano.api.v1 import request_statistics
from murano.common.i18n import _
from murano.common import policy
from murano.common import utils
from murano.common import wsgi
from murano.db import models
from murano.db.services import core_services
from murano.db.services import environment_templates as env_temps
from murano.db.services import environments as envs
from murano.db.services import sessions

LOG = logging.getLogger(__name__)

API_NAME = 'Templates'


class Controller(object):
    @request_statistics.stats_count(API_NAME, 'Index')
    def index(self, request):
        """Lists the env templates associated to an tenant-id

        It lists the env templates associated to an tenant-id.
        :param request: The operation request.
        :return: the env template description list.
        """
        LOG.debug('EnvTemplates:List')
        policy.check('list_env_templates', request.context)
        tenant_id = request.context.tenant
        filters = {}
        if request.GET.get('is_public'):
            is_public = request.GET.get('is_public', 'false').lower() == 'true'
            if not is_public:

                filters['is_public'] = False
                filters['tenant_id'] = tenant_id
            elif is_public:
                filters['is_public'] = True

            list_templates = env_temps.EnvTemplateServices.\
                get_env_templates_by(filters)

        else:
            filters = (models.EnvironmentTemplate.is_public,
                       models.EnvironmentTemplate.tenant_id == tenant_id)
            list_templates = env_temps.EnvTemplateServices.\
                get_env_templates_or_by(filters)

        list_templates = [temp.to_dict() for temp in list_templates]
        return {"templates": list_templates}

    @request_statistics.stats_count(API_NAME, 'Create')
    def create(self, request, body):
        """Creates the env template from the payload

        This payload can contain just the template name, or include
        also service information.
        :param request: the operation request.
        :param body: the env template description
        :return: the description of the created template.
        """
        LOG.debug('EnvTemplates:Create <Body {body}>'.format(body=body))
        policy.check('create_env_template', request.context)

        self._validate_body_name(body)
        try:
            LOG.debug('ENV TEMP NAME: {templ_name}>'.
                      format(templ_name=body['name']))
            template = env_temps.EnvTemplateServices.create(
                body.copy(), request.context.tenant)
            return template.to_dict()
        except db_exc.DBDuplicateEntry:
            msg = _('Env Template with specified name already exists')
            LOG.exception(msg)
            raise exc.HTTPConflict(msg)

    @request_statistics.stats_count(API_NAME, 'Show')
    def show(self, request, env_template_id):
        """It shows the description of a template

        :param request: the operation request.
        :param env_template_id: the env template ID.
        :return: the description of the env template.
        """
        LOG.debug('Templates:Show <Id: {templ_id}>'.format(
            templ_id=env_template_id))
        target = {"env_template_id": env_template_id}
        policy.check('show_env_template', request.context, target)

        self._validate_request(request, env_template_id)

        template = env_temps.EnvTemplateServices.\
            get_env_template(env_template_id)
        temp = template.to_dict()

        get_data = core_services.CoreServices.get_template_data
        temp['services'] = get_data(env_template_id, '/services')
        return temp

    @request_statistics.stats_count(API_NAME, 'Update')
    def update(self, request, env_template_id, body):
        """It updates the description template

        :param request: the operation request.
        :param env_template_id: the env template ID.
        :param body: the description to be updated
        :return: the updated template description.
        """
        LOG.debug('Templates:Update <Id: {templ_id}, '
                  'Body: {body}>'.format(templ_id=env_template_id, body=body))
        target = {"env_template_id": env_template_id}
        policy.check('update_env_template', request.context, target)

        self._validate_request(request, env_template_id)
        try:
            LOG.debug('ENV TEMP NAME: {temp_name}>'.format(
                temp_name=body['name']))
            if not str(body['name']).strip():
                msg = _('Environment Template must contain at least one '
                        'non-white space symbol')
                LOG.exception(msg)
                raise exc.HTTPBadRequest(msg)
        except Exception:
            msg = _('EnvTemplate body is incorrect')
            LOG.exception(msg)
            raise exc.HTTPBadRequest(msg)

        template = env_temps.EnvTemplateServices.update(env_template_id, body)
        return template.to_dict()

    @request_statistics.stats_count(API_NAME, 'Delete')
    def delete(self, request, env_template_id):
        """It deletes the env template

        :param request: the operation request.
        :param env_template_id: the template ID.
        """
        LOG.debug('EnvTemplates:Delete <Id: {templ_id}>'.format(
            templ_id=env_template_id))
        target = {"env_template_id": env_template_id}
        policy.check('delete_env_template', request.context, target)
        self._validate_request(request, env_template_id)
        env_temps.EnvTemplateServices.delete(env_template_id)
        env_temps.EnvTemplateServices.remove(env_template_id)
        return

    def has_services(self, template):
        """"It checks if the template has services

        :param template: the template to check.
        :return: True or False
        """
        if not template.description:
            return False

        if (template.description.get('services')):
            return True
        return False

    @request_statistics.stats_count(API_NAME, 'Create_environment')
    def create_environment(self, request, env_template_id, body):
        """Creates environment and session from template

        :param request: operation request
        :param env_template_id: environment template ID
        :param body: the environment name
        :return: session_id and environment_id
        """
        target = {"env_template_id": env_template_id}
        policy.check('create_environment', request.context, target)

        self._validate_request(request, env_template_id)
        LOG.debug('Templates:Create environment <Id: {templ_id}>'.
                  format(templ_id=env_template_id))
        template = env_temps.EnvTemplateServices.\
            get_env_template(env_template_id)
        self._validate_body_name(body)
        try:
            environment = envs.EnvironmentServices.create(
                body.copy(), request.context)
        except db_exc.DBDuplicateEntry:
            msg = _('Environment with specified name already exists')
            LOG.exception(msg)
            raise exc.HTTPConflict(explanation=msg)

        user_id = request.context.user
        session = sessions.SessionServices.create(environment.id, user_id)

        if self.has_services(template):
            services_node = utils.TraverseHelper.get("services",
                                                     template.description)
            utils.TraverseHelper.update("/Objects/services",
                                        services_node,
                                        environment.description)

        envs.EnvironmentServices.save_environment_description(
            session.id,
            environment.description,
            inner=False
        )
        return {"session_id": session.id, "environment_id": environment.id}

    @request_statistics.stats_count(API_NAME, 'Clone')
    def clone(self, request, env_template_id, body):
        """Clones env template from another tenant

        It clones the env template from another env template
        from other tenant.
        :param request: the operation request.
        :param env_template_id: the env template ID.
        :param body: the request body.
        :return: the description of the created template.
        """

        LOG.debug('EnvTemplates:Clone <Env Template {1} for body {0}>'.
                  format(body, env_template_id))
        policy.check('clone_env_template', request.context)

        old_env_template = self._validate_exists(env_template_id)

        if not old_env_template.get('is_public'):
            msg = _('User has no access to these resources.')
            LOG.error(msg)
            raise exc.HTTPForbidden(explanation=msg)
        self._validate_body_name(body)
        LOG.debug('ENV TEMP NAME: {0}'.format(body['name']))

        try:
            is_public = body.get('is_public', False)
            template = env_temps.EnvTemplateServices.clone(
                env_template_id, request.context.tenant, body['name'],
                is_public)
        except db_exc.DBDuplicateEntry:
            msg = _('Env template with specified name already exists')
            LOG.error(msg)
            raise exc.HTTPConflict(explanation=msg)

        return template.to_dict()

    def _validate_request(self, request, env_template_id):
        env_template = self._validate_exists(env_template_id)
        if env_template.is_public or request.context.is_admin:
            return
        if env_template.tenant_id != request.context.tenant:
            msg = _('User has no access to these resources.')
            LOG.error(msg)
            raise exc.HTTPForbidden(explanation=msg)

    def _validate_exists(self, env_template_id):
        env_template_exists = env_temps.EnvTemplateServices.env_template_exist
        if not env_template_exists(env_template_id):
            msg = _('EnvTemplate <TempId {temp_id}> is not found').format(
                temp_id=env_template_id)
            LOG.error(msg)
            raise exc.HTTPNotFound(explanation=msg)
        get_env_template = env_temps.EnvTemplateServices.get_env_template
        return get_env_template(env_template_id)

    def _validate_body_name(self, body):

        if not('name' in body and body['name'].strip()):
            msg = _('Please, specify a name of the environment template.')
            LOG.error(msg)
            raise exc.HTTPBadRequest(explanation=msg)

        name = str(body['name'])
        if len(name) > 255:
            msg = _('Environment template name should be 255 characters '
                    'maximum')
            LOG.error(msg)
            raise exc.HTTPBadRequest(explanation=msg)


def create_resource():
    return wsgi.Resource(Controller())



#    Copyright (c) 2014 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from http import client as http_client
from oslo_serialization import jsonutils
import webob.dec

from murano.common import wsgi


class Controller(object):

    """A wsgi controller that reports which API versions are supported."""

    def index(self, req):
        """Respond to a request for all OpenStack API versions."""
        def build_version_object(version, path, status):
            return {
                'id': 'v%s' % version,
                'status': status,
                'links': [
                    {
                        'rel': 'self',
                        'href': '%s/%s/' % (req.host_url, path),
                    },
                ],
            }

        version_objs = []
        version_objs.extend([
            build_version_object(1.0, 'v1', 'CURRENT'),
        ])

        response = webob.Response(request=req,
                                  status=http_client.MULTIPLE_CHOICES,
                                  content_type='application/json')
        response.text = jsonutils.dumps(dict(versions=version_objs))
        return response

    @webob.dec.wsgify(RequestClass=wsgi.Request)
    def __call__(self, req):
        return self.index(req)


def create_resource(conf):
    return wsgi.Resource(Controller())



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from alembic import context
from sqlalchemy import create_engine, pool

from murano.db import models


# this is the Alembic Config object, which provides
# access to the values within the .ini file in use.
config = context.config
murano_config = config.murano_config

# add your model's MetaData object here
# for 'autogenerate' support
# from myapp import mymodel
target_metadata = models.Base.metadata


def run_migrations_online():
    """Run migrations in 'online' mode.

    In this scenario we need to create an Engine
    and associate a connection with the context.

    """
    engine = create_engine(
        murano_config.database.connection,
        poolclass=pool.NullPool)

    with engine.connect() as connection:
        context.configure(connection=connection,
                          target_metadata=target_metadata)
        with context.begin_transaction():
            context.run_migrations()


run_migrations_online()



#    Copyright (c) 2015 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


import os

from oslo_config import cfg
from oslo_log import log as logging
from paste import deploy

from murano.common.i18n import _

CONF = cfg.CONF


def _get_deployment_flavor():
    """Retrieve the paste_deploy.flavor config item

       Retrieve the paste_deploy.flavor config item, formatted appropriately
       for appending to the application name.
    """
    flavor = CONF.paste_deploy.flavor
    return '' if not flavor else ('-' + flavor)


def _get_paste_config_path():
    paste_suffix = '-paste.ini'
    conf_suffix = '.conf'
    if CONF.config_file:
        # Assume paste config is in a paste.ini file corresponding
        # to the last config file
        path = CONF.config_file[-1].replace(conf_suffix, paste_suffix)
    else:
        path = CONF.prog + '-paste.ini'
    return CONF.find_file(os.path.basename(path))


def _get_deployment_config_file():
    """Retrieve the deployment_config_file config item

       Retrieve the deployment_config_file config item, formatted as an
       absolute pathname.
    """
    path = CONF.paste_deploy.config_file
    if not path:
        path = _get_paste_config_path()
    if not path:
        msg = _("Unable to locate paste config file for %s.") % CONF.prog
        raise RuntimeError(msg)
    return os.path.abspath(path)


def load_paste_app(app_name=None):
    """Builds and returns a WSGI app from a paste config file.

    We assume the last config file specified in the supplied ConfigOpts
    object is the paste config file.

    :param app_name: name of the application to load

    :raises RuntimeError when config file cannot be located or application
            cannot be loaded from config file
    """
    if app_name is None:
        app_name = CONF.prog

    # append the deployment flavor to the application name,
    # in order to identify the appropriate paste pipeline
    app_name += _get_deployment_flavor()

    conf_file = _get_deployment_config_file()

    try:
        logger = logging.getLogger(__name__)
        logger.debug("Loading {app_name} from {conf_file}".format(
            conf_file=conf_file, app_name=app_name))

        app = deploy.loadapp("config:%s" % conf_file, name=app_name)

        return app
    except (LookupError, ImportError) as e:
        msg = _("Unable to load %(app_name)s from configuration file"
                " %(conf_file)s. \nGot: %(e)r") % {'conf_file': conf_file,
                                                   'app_name': app_name,
                                                   'e': e}
        logger.error(msg)
        raise RuntimeError(msg)



# Copyright (c) 2016 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy

from murano.tests.unit.dsl.foundation import object_model as om
from murano.tests.unit.dsl.foundation import test_case


class TestObjectsCopyMerge(test_case.DslTestCase):
    def test_merged(self):
        gen1_model = om.Object(
            'TestObjectsCopyMergeSampleClass', 'rootNode',
            value='node1',
            nodes=[
                om.Object('TestObjectsCopyMergeSampleClass',
                          value='node2',
                          nodes=[om.Ref('rootNode')])
            ])

        gen2_model = copy.deepcopy(gen1_model)
        gen2_model['nodes'] = []
        gen2_model['value'] = 'node1-updated'

        model = {
            'Objects': gen2_model,
            'ObjectsCopy': gen1_model
        }

        runner = self.new_runner(model)
        self.assertEqual(['node2', 'node1-updated'], self.traces)
        self.assertEqual('It works!', runner.testMethod())
